/*
 *   HP8x.c
 *
 *   Copyright 2004 Olivier De Smet
 */

//
// Main file for HP8x emulator
//

#include "StdAfx.h"
#include "HP8x.h"
#include "resource.h"
#include "kml.h"
#include "debugger.h"

#define VERSION   "1.00beta2"

#if defined DEBUG_ONDISK
HANDLE hDebug;
#endif

HANDLE hHeap;								// private heap

#ifdef _DEBUG
LPTSTR szNoTitle = _T("HP8x ")_T(VERSION)_T(" Debug");
#else
LPTSTR szNoTitle = _T("HP8x ")_T(VERSION);
#endif
// LPTSTR szAppName = _T("HP8x");				// application name for DDE server
// LPTSTR szTopic   = _T("Nothing");			// topic for DDE server
LPTSTR szTitle   = NULL;

static const LPCTSTR szLicence =
	_T("This program is free software; you can redistribute it and/or modify\r\n")
	_T("it under the terms of the GNU General Public License as published by\r\n")
	_T("the Free Software Foundation; either version 2 of the License, or\r\n")
	_T("(at your option) any later version.\r\n")
	_T("\r\n")
	_T("This program is distributed in the hope that it will be useful,\r\n")
	_T("but WITHOUT ANY WARRANTY; without even the implied warranty of\r\n")
	_T("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\r\n")
	_T("See the GNU General Public License for more details.\r\n")
	_T("\r\n")
	_T("You should have received a copy of the GNU General Public License\r\n")
	_T("along with this program; if not, write to the Free Software Foundation,\r\n")
    _T("Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA");

static LPCTSTR          *ppArgv;					// command line arguments
static INT              nArgc;						// no. of command line arguments
static DWORD            lThreadId;
static LARGE_INTEGER    lFreq;						// high performance counter frequency

CRITICAL_SECTION csGDILock;					// critical section for hWindowDC
CRITICAL_SECTION csKeyLock;					// critical section for key scan
HANDLE           hThread;
HANDLE           hEventShutdn;				// event handle to stop cpu thread

HINSTANCE        hApp = NULL;
HWND             hWnd = NULL;
HWND             hDlgDebug = NULL;			// handle for debugger dialog
HWND             hDlgFind = NULL;			// handle for debugger find dialog
HDC              hWindowDC = NULL;
HPALETTE         hPalette = NULL;
HPALETTE         hOldPalette = NULL;		// old palette of hWindowDC
BOOL             bAutoSave = FALSE;
BOOL             bAutoSaveOnExit = FALSE;
BOOL             bAlwaysDisplayLog = TRUE;

static int		 nDumpNr = 0;


//################
//#
//#    Window Status
//#
//################

VOID SetWindowTitle(LPTSTR szString)
{
	if (szTitle) HeapFree(hHeap, 0, szTitle);

	_ASSERT(hWnd != NULL);
	if (szString)
	{
		szTitle = DuplicateString(szString);
		SetWindowText(hWnd, szTitle);
	}
	else
	{
		szTitle = NULL;
		SetWindowText(hWnd, szNoTitle);
	}
}

VOID UpdateWindowStatus(VOID)
{
	if (hWnd)								// window open
	{
		BOOL bRun         = nState == SM_RUN;
		UINT uRun         = bRun                   ? MF_ENABLED : MF_GRAYED;
		UINT uBackup      = bBackup                ? MF_ENABLED : MF_GRAYED;		
		UINT uHP9121	  = (Chipset.iop[7].iop == 1)  ? MF_ENABLED : MF_GRAYED;
		UINT uHP9121_0	  = (Chipset.iop[7].iop == 1) && (Chipset.D70x_disk[0][0] != 0) ? MF_ENABLED : MF_GRAYED;
		UINT uHP9121_1	  = (Chipset.iop[7].iop == 1) && (Chipset.D70x_disk[1][0] != 0) ? MF_ENABLED : MF_GRAYED;
		UINT uHP9121_2	  = (Chipset.iop[7].iop == 1) && (Chipset.D70x_disk[2][0] != 0) ? MF_ENABLED : MF_GRAYED;
		UINT uHP9121_3	  = (Chipset.iop[7].iop == 1) && (Chipset.D70x_disk[3][0] != 0) ? MF_ENABLED : MF_GRAYED;
#if defined HP87
		UINT uHP9121_4	  = (Chipset.iop[7].iop == 1) && (Chipset.D70x_disk[4][0] != 0) ? MF_ENABLED : MF_GRAYED;
		UINT uHP9121_5	  = (Chipset.iop[7].iop == 1) && (Chipset.D70x_disk[5][0] != 0) ? MF_ENABLED : MF_GRAYED;
		UINT uHP9121_6	  = (Chipset.iop[7].iop == 1) && (Chipset.D70x_disk[6][0] != 0) ? MF_ENABLED : MF_GRAYED;
		UINT uHP9121_7	  = (Chipset.iop[7].iop == 1) && (Chipset.D70x_disk[7][0] != 0) ? MF_ENABLED : MF_GRAYED;
		UINT uHP9121_8	  = (Chipset.iop[7].iop == 1) && (Chipset.D70x_disk[8][0] != 0) ? MF_ENABLED : MF_GRAYED;
		UINT uHP9121_9	  = (Chipset.iop[7].iop == 1) && (Chipset.D70x_disk[9][0] != 0) ? MF_ENABLED : MF_GRAYED;
#endif
		HMENU hMenu = GetMenu(hWnd);		// get menu handle

		EnableMenuItem(hMenu, ID_FILE_NEW, MF_ENABLED);
		EnableMenuItem(hMenu, ID_FILE_OPEN, MF_ENABLED);
		EnableMenuItem(hMenu, ID_FILE_SAVE, (bRun && szCurrentFilename[0]) ? MF_ENABLED : MF_GRAYED);
		EnableMenuItem(hMenu, ID_FILE_SAVEAS, uRun);
		EnableMenuItem(hMenu, ID_FILE_CLOSE, uRun);
		EnableMenuItem(hMenu, ID_D700_LOAD, uHP9121);
		EnableMenuItem(hMenu, ID_D701_LOAD, uHP9121);
		EnableMenuItem(hMenu, ID_D702_LOAD, uHP9121);
		EnableMenuItem(hMenu, ID_D703_LOAD, uHP9121);
#if defined HP87
		EnableMenuItem(hMenu, ID_D704_LOAD, uHP9121);
		EnableMenuItem(hMenu, ID_D705_LOAD, uHP9121);
		EnableMenuItem(hMenu, ID_D706_LOAD, uHP9121);
		EnableMenuItem(hMenu, ID_D707_LOAD, uHP9121);
		EnableMenuItem(hMenu, ID_D708_LOAD, uHP9121);
		EnableMenuItem(hMenu, ID_D709_LOAD, uHP9121);
#endif
		EnableMenuItem(hMenu, ID_D700_SAVE, uHP9121_0);
		EnableMenuItem(hMenu, ID_D701_SAVE, uHP9121_1);
		EnableMenuItem(hMenu, ID_D702_SAVE, uHP9121_2);
		EnableMenuItem(hMenu, ID_D703_SAVE, uHP9121_3);
		
#if defined HP87
		EnableMenuItem(hMenu, ID_D704_SAVE, uHP9121_4);
		EnableMenuItem(hMenu, ID_D705_SAVE, uHP9121_5);
		EnableMenuItem(hMenu, ID_D706_SAVE, uHP9121_6);
		EnableMenuItem(hMenu, ID_D707_SAVE, uHP9121_7);
		EnableMenuItem(hMenu, ID_D708_SAVE, uHP9121_8);
		EnableMenuItem(hMenu, ID_D709_SAVE, uHP9121_9);
#endif
		EnableMenuItem(hMenu, ID_D700_EJECT, uHP9121_0);
		EnableMenuItem(hMenu, ID_D701_EJECT, uHP9121_1);
		EnableMenuItem(hMenu, ID_D702_EJECT, uHP9121_2);
		EnableMenuItem(hMenu, ID_D703_EJECT, uHP9121_3);
#if defined HP87
		EnableMenuItem(hMenu, ID_D704_EJECT, uHP9121_4);
		EnableMenuItem(hMenu, ID_D705_EJECT, uHP9121_5);
		EnableMenuItem(hMenu, ID_D706_EJECT, uHP9121_6);
		EnableMenuItem(hMenu, ID_D707_EJECT, uHP9121_7);
		EnableMenuItem(hMenu, ID_D708_EJECT, uHP9121_8);
		EnableMenuItem(hMenu, ID_D709_EJECT, uHP9121_9);
#endif
		EnableMenuItem(hMenu, ID_VIEW_COPY, uRun);
		EnableMenuItem(hMenu, ID_EDIT_PASTE, uRun);
		EnableMenuItem(hMenu, ID_VIEW_RESET, uRun);
		EnableMenuItem(hMenu, ID_BACKUP_SAVE, uRun);
		EnableMenuItem(hMenu, ID_BACKUP_RESTORE, uBackup);
		EnableMenuItem(hMenu, ID_BACKUP_DELETE, uBackup);
//		EnableMenuItem(hMenu, ID_TOOL_DISASM, uRun);
		EnableMenuItem(hMenu, ID_TOOL_DEBUG, (bRun && nDbgState == DBG_OFF) ? MF_ENABLED : MF_GRAYED);
	}
}

//################
//#
//#    Clipboard Tool
//#
//################

static VOID CopyItemsToClipboard(HWND hWnd)		// save selected Listbox Items to Clipboard
{
	LONG  i;
	LPINT lpnCount;

	// get number of selections
	if ((i = SendMessage(hWnd,LB_GETSELCOUNT,0,0)) == 0)
		return;								// no items selected

	if ((lpnCount = HeapAlloc(hHeap,0,i * sizeof(INT))) != NULL)
	{
		HANDLE hClipObj;
		LONG j,lMem = 0;

		// get indexes of selected items
		i = SendMessage(hWnd,LB_GETSELITEMS,i,(LPARAM) lpnCount);
		for (j = 0;j < i;++j)				// scan all selected items
		{
			// calculate total amount of needed memory
			lMem += SendMessage(hWnd,LB_GETTEXTLEN,lpnCount[j],0) + 2;
		}
		// allocate clipboard data
		if ((hClipObj = GlobalAlloc(GMEM_MOVEABLE,lMem + 1)) != NULL)
		{
			LPBYTE lpData;

			if ((lpData = GlobalLock(hClipObj)))
			{
				for (j = 0;j < i;++j)		// scan all selected items
				{
					#if defined _UNICODE
					{
						INT nLength = SendMessage(hWnd,LB_GETTEXTLEN,lpnCount[j],0) + 1;

						LPTSTR szTmp = HeapAlloc(hHeap,0,nLength * sizeof(szTmp[0]));
						if (szTmp != NULL)
						{
							SendMessage(hWnd,LB_GETTEXT,lpnCount[j],(LPARAM) szTmp);
							WideCharToMultiByte(CP_ACP, WC_COMPOSITECHECK,
												szTmp, nLength,
												lpData, nLength, NULL, NULL);
							HeapFree(hHeap,0,szTmp);
							lpData += nLength - 1;
						}
					}
					#else
					{
						lpData += SendMessage(hWnd,LB_GETTEXT,lpnCount[j],(LPARAM) lpData);
					}
					#endif
					*lpData++ = '\r';
					*lpData++ = '\n';
				}
				*lpData = 0;				// set end of string
				GlobalUnlock(hClipObj);		// unlock memory
			}

			if (OpenClipboard(hWnd))
			{
				if (EmptyClipboard())
					SetClipboardData(CF_TEXT,hClipObj);
				else
					GlobalFree(hClipObj);
				CloseClipboard();
			}
			else							// clipboard open failed
			{
				GlobalFree(hClipObj);
			}
		}
		HeapFree(hHeap,0,lpnCount);			// free item table
	}
}

//################
//#
//#    Settings
//#
//################

static BOOL CALLBACK SettingsProc(HWND hDlg, UINT message, DWORD wParam, LONG lParam)
{
	switch (message)
	{
	case WM_INITDIALOG:
		// init speed checkbox
		CheckDlgButton(hDlg, IDC_DATE, bSetDate);
		CheckDlgButton(hDlg, IDC_SOUND, bSound);
		CheckDlgButton(hDlg, IDC_REALSPEED, bRealSpeed);
		CheckDlgButton(hDlg, IDC_AUTOSAVE, bAutoSave);
		CheckDlgButton(hDlg, IDC_AUTOSAVEONEXIT, bAutoSaveOnExit);
		CheckDlgButton(hDlg, IDC_ALWAYSDISPLOG, bAlwaysDisplayLog);
		return TRUE;
	case WM_COMMAND:
		if (wParam == IDOK)
		{
			// get speed checkbox value
			bSetDate = IsDlgButtonChecked(hDlg, IDC_DATE);
			bSound = IsDlgButtonChecked(hDlg, IDC_SOUND);
			bRealSpeed = IsDlgButtonChecked(hDlg, IDC_REALSPEED);
			bAutoSave = IsDlgButtonChecked(hDlg, IDC_AUTOSAVE);
			bAutoSaveOnExit = IsDlgButtonChecked(hDlg, IDC_AUTOSAVEONEXIT);
			bAlwaysDisplayLog = IsDlgButtonChecked(hDlg, IDC_ALWAYSDISPLOG);
			SetSpeed(bRealSpeed);			// set speed
			CloseWaves();					// set sound
			if (bSound)	
				bSound = InitWaves();
			EndDialog(hDlg, wParam);
		}
		if (wParam == IDCANCEL)
		{
			EndDialog(hDlg, wParam);
		}
		break;
	} 
	return FALSE;
    UNREFERENCED_PARAMETER(lParam);
}


//################
//#
//#    Save Helper
//#
//################

static UINT SaveChanges(BOOL bAuto)
{
	UINT uReply;

	if (pbyRom == NULL) return IDNO;

	if (bAuto)
		uReply = IDYES;
	else
		uReply = YesNoCancelMessage(_T("Do you want to save changes ?"));

	if (uReply != IDYES) return uReply;

	if (szCurrentFilename[0]==0)
	{ // Save As...
		uReply = GetSaveAsFilename();
		if (uReply != IDOK) return uReply;
		if (!SaveDocumentAs(szBufferFilename)) return IDCANCEL;
		WriteLastDocument(szCurrentFilename);
		return IDYES;
	}

	SaveDocument();
	return IDYES;
}

//################
//#
//#    Message Handlers
//#
//################

static LRESULT OnCreate(HWND hWindow)
{
	InitializeCriticalSection(&csGDILock);
	InitializeCriticalSection(&csKeyLock);

	hWnd = hWindow;
	hWindowDC = GetDC(hWnd);
	DragAcceptFiles(hWnd,FALSE);				// no support for dropped files
	return 0;
}

static LRESULT OnDestroy(HWND hWindow)
{
	DragAcceptFiles(hWnd,FALSE);			// no WM_DROPFILES message any more
	SwitchToState(SM_RETURN);				// exit emulation thread
	SetWindowTitle(NULL);					// free memory of title
	ReleaseDC(hWnd, hWindowDC);
	hWindowDC = NULL;						// hWindowDC isn't valid any more
	hWnd = NULL;

	DeleteCriticalSection(&csGDILock);
	DeleteCriticalSection(&csKeyLock);

	PostQuitMessage(0);					// exit message loop
	return 0;
	UNREFERENCED_PARAMETER(hWindow);
}

static LRESULT OnPaint(HWND hWindow)
{
	PAINTSTRUCT Paint;
	HDC hPaintDC;

	hPaintDC = BeginPaint(hWindow, &Paint);
	if (hMainDC != NULL)
	{
		RECT rcMainPaint = Paint.rcPaint;
		rcMainPaint.left   += nBackgroundX;	// coordinates in source bitmap
		rcMainPaint.top    += nBackgroundY;
		rcMainPaint.right  += nBackgroundX;
		rcMainPaint.bottom += nBackgroundY;

		EnterCriticalSection(&csGDILock);	// solving NT GDI problems
		{
			// redraw background bitmap
			BitBlt(hPaintDC, Paint.rcPaint.left, Paint.rcPaint.top,
				   Paint.rcPaint.right-Paint.rcPaint.left, Paint.rcPaint.bottom-Paint.rcPaint.top,
				   hMainDC, rcMainPaint.left, rcMainPaint.top, SRCCOPY);
			// redraw main display area
			BitBlt(hWindowDC, nLcdX, nLcdY,
				   nLcdCol[nLcdxZoom],nLcdRow[nLcdyZoom],
				   hLcdDC, 0, 0, SRCCOPY);
			GdiFlush();
		}
		LeaveCriticalSection(&csGDILock); 
		UpdateAnnunciators(); 
		RefreshButtons(&rcMainPaint);
	}
	EndPaint(hWindow, &Paint);
	return 0;
}

//
// ID_FILE_NEW
//
static LRESULT OnFileNew(VOID)
{
	SaveBackup();
	if (pbyRom)
	{
		SwitchToState(SM_INVALID);
		if (IDCANCEL == SaveChanges(bAutoSave))
			goto cancel;
	}
	if (NewDocument()) SetWindowTitle(_T("Untitled"));
	UpdateWindowStatus();
cancel:
	if (pbyRom) SwitchToState(SM_RUN);
	return 0;
}

//
// ID_FILE_OPEN
//
static LRESULT OnFileOpen(VOID)
{
	if (pbyRom)
	{
		SwitchToState(SM_INVALID);
		if (IDCANCEL == SaveChanges(bAutoSave))
			goto cancel;
	}
	if (GetOpenFilename())
	{
		OpenDocument(szBufferFilename);
	}
cancel:
	if (pbyRom) SwitchToState(SM_RUN);
	return 0;
}

//
// ID_FILE_SAVE
//
static LRESULT OnFileSave(VOID)
{
	if (pbyRom == NULL) return 0;
	SwitchToState(SM_INVALID);
	SaveChanges(TRUE);
	SwitchToState(SM_RUN);
	return 0;
}

//
// ID_FILE_SAVEAS
//
static LRESULT OnFileSaveAs(VOID)
{
	UINT uReply;

	if (pbyRom == NULL) return 0;
	SwitchToState(SM_INVALID);

	uReply = GetSaveAsFilename();
	if (uReply != IDOK)
	{
		SwitchToState(SM_RUN);
		return 0;
	}
	if (!SaveDocumentAs(szBufferFilename))
	{
		SwitchToState(SM_RUN);
		return 0;
	}
	WriteLastDocument(szCurrentFilename);

	SwitchToState(SM_RUN);
	return 0;
}

//
// ID_FILE_CLOSE
//
static LRESULT OnFileClose(VOID)
{
	if (pbyRom == NULL) return 0;
	SwitchToState(SM_INVALID);
	if (SaveChanges(bAutoSave)!=IDCANCEL)
	{
		DisableDebugger();
		KillKML();
		ResetDocument();
		SetWindowTitle(NULL);
	}
	else
	{
		SwitchToState(SM_RUN);
	}
	return 0;
}

//
// ID_FILE_EXIT
//
// WM_SYS_CLOSE
//
static LRESULT OnFileExit(VOID)
{
	SwitchToState(SM_INVALID);				// hold emulation thread
	if (SaveChanges(bAutoSaveOnExit) == IDCANCEL)
	{
		SwitchToState(SM_RUN);				// on cancel restart emulation thread
		return 0;
	} 
	DestroyWindow(hWnd); 
	return 0;
}

//
// ID_COPY_STRING
//
static LRESULT OnCopyString(VOID)			// copy data from screen (ALPHA)
{
	HANDLE hClipObj;
	LPBYTE lpData, lpAddress;
	DWORD dwSize;
	BOOL bCrlf = FALSE;							// flag to treat 0x0D -> CR LF (only once)

	_ASSERT(nState == SM_RUN);				// emulator must be in RUN state
	if (WaitForIdleState())					// wait for basic IDLE state
	{
		InfoMessage(_T("The emulator is busy."));
		return 0;
	}

#if defined HP87
	if ((Chipset.crtsts & 0x80) == 0x80)	// GRAPH mode
#endif
#if defined HP85
	if ((Chipset.crtwsts & 0x80) == 0x80)	// GRAPH mode
#endif
	{
		InfoMessage(_T("The emulator is in GRAPH mode."));
		return 0;
	}

	lpAddress = Chipset.Video;									// address of screen
#if defined HP87
	dwSize = ((Chipset.crtsts & 0x40) == 0x40) ? 0x3FC0:0x10E0;	// length of screen (ALL or normal)
#endif
#if defined HP85
	dwSize = 0x0800;
#endif
	// memory allocation for clipboard data
	if ((hClipObj = GlobalAlloc(GMEM_MOVEABLE,0x4000 + 205)) == NULL)	// take care of 204 0x0D -> 204 CR LF
		goto error;

	if ((lpData = GlobalLock(hClipObj)))	// lock memory
	{
		// copy data into clipboard buffer
		for ( ;dwSize > 0; dwSize--)
		{
			switch(*lpAddress)
			{
			case 0x00:
				break;
			case 0x0D:
				if (!bCrlf)
				{
					*lpData++ = 0x0D;
					*lpData++ = 0x0A;
					bCrlf = TRUE;
				}
				break;
			default:
				bCrlf = FALSE;
				*lpData++ = *lpAddress;
				break;
			}
			lpAddress++;
		}
		*lpData = 0;						// set end of string

		GlobalUnlock(hClipObj);				// unlock memory

		if (OpenClipboard(hWnd))
		{
			if (EmptyClipboard())
				SetClipboardData(CF_TEXT,hClipObj);
			else
				GlobalFree(hClipObj);
			CloseClipboard();
		}
	}
	else									// lock memory failed
	{
		GlobalFree(hClipObj);
	}

error:
	return 0;
}

//
// ID_PASTE_STRING
//
static LRESULT OnPasteString(VOID)			// paste data to keyboard
{
	HANDLE hClipObj;
	LPBYTE lpClipdata, lpData;

	SuspendDebugger();						// suspend debugger
	bDbgAutoStateCtrl = FALSE;				// disable automatic debugger state control

	_ASSERT(nState == SM_RUN);				// emulator must be in RUN state

	if (WaitForIdleState())					// wait for basic idle state
	{
		InfoMessage(_T("The emulator is busy."));
		goto cancel;
	}

	if (OpenClipboard(hWnd))
	{
		if (   IsClipboardFormatAvailable(CF_TEXT)
			|| IsClipboardFormatAvailable(CF_OEMTEXT))
		{
			if ((hClipObj = GetClipboardData(CF_TEXT)))
			{
				if ((lpClipdata = GlobalLock(hClipObj)))
				{
					DWORD dwSize = strlen(lpClipdata);
					if ((lpData = (LPBYTE) HeapAlloc(hHeap, 0, dwSize)))
					{
						memcpy(lpData,lpClipdata,dwSize);			// copy data
						WriteToKeyboard(lpData,dwSize);				// do the copy
						HeapFree(hHeap, 0, lpData);
					}
					GlobalUnlock(hClipObj);
				}
			}
		}
		else								// unknown clipboard format
		{
			MessageBeep(MB_OK);				// error beep
		}
		CloseClipboard();
	}

cancel:
	bDbgAutoStateCtrl = TRUE;				// enable automatic debugger state control
	ResumeDebugger();
	return 0;
}

//
// ID_COPY_SCREEN
//
static LRESULT OnCopyScreen(VOID)
{
	if (OpenClipboard(hWnd))
	{
		if (EmptyClipboard())
		{
			// DIB bitmap
			#define WIDTHBYTES(bits) (((bits) + 31) / 32 * 4)
			#define PALVERSION       0x300

			BITMAP bm;
			LPBITMAPINFOHEADER lpbi;
			PLOGPALETTE ppal;
			HBITMAP hBmp;
			HDC hBmpDC;
			HANDLE hClipObj;
			WORD wBits;
			DWORD dwLen, dwSizeImage;

			hBmp = CreateCompatibleBitmap(hLcdDC, nLcdCol[nLcdxZoom], nLcdRow[nLcdyZoom]);
			hBmpDC = CreateCompatibleDC(hLcdDC);
			hBmp = SelectObject(hBmpDC, hBmp);
			BitBlt(hBmpDC, 0, 0, nLcdCol[nLcdxZoom], nLcdRow[nLcdyZoom], hLcdDC, 0, 0, SRCCOPY);
			hBmp = SelectObject(hBmpDC, hBmp);

			// fill BITMAP structure for size information
			GetObject(hBmp, sizeof(bm), &bm);

			wBits = bm.bmPlanes * bm.bmBitsPixel;
			// make sure bits per pixel is valid
			if (wBits <= 1)
				wBits = 1;
			else if (wBits <= 4)
				wBits = 4;
			else if (wBits <= 8)
				wBits = 8;
			else // if greater than 8-bit, force to 24-bit
				wBits = 24;

			dwSizeImage = WIDTHBYTES((DWORD)bm.bmWidth * wBits) * bm.bmHeight;

			// calculate memory size to store CF_DIB data
			dwLen = sizeof(BITMAPINFOHEADER) + dwSizeImage;
			if (wBits != 24)				// a 24 bitcount DIB has no color table
			{
				// add size for color table
				dwLen += (1 << wBits) * sizeof(RGBQUAD);
			}

			// memory allocation for clipboard data
			if ((hClipObj = GlobalAlloc(GMEM_MOVEABLE, dwLen)) != NULL)
			{
				lpbi = GlobalLock(hClipObj);
				// initialize BITMAPINFOHEADER
				lpbi->biSize = sizeof(BITMAPINFOHEADER);
				lpbi->biWidth = bm.bmWidth;
				lpbi->biHeight = bm.bmHeight;
				lpbi->biPlanes = 1;
				lpbi->biBitCount = wBits;
				lpbi->biCompression = BI_RGB;
				lpbi->biSizeImage = dwSizeImage;
				lpbi->biXPelsPerMeter = 0;
				lpbi->biYPelsPerMeter = 0;
				lpbi->biClrUsed = 0;
				lpbi->biClrImportant = 0;
				// get bitmap color table and bitmap data
				GetDIBits(hBmpDC, hBmp, 0, lpbi->biHeight, (LPBYTE)lpbi + dwLen - dwSizeImage,
                          (LPBITMAPINFO)lpbi, DIB_RGB_COLORS);
				GlobalUnlock(hClipObj);
				SetClipboardData(CF_DIB, hClipObj);

				// get number of entries in the logical palette
				GetObject(hPalette,sizeof(WORD),&wBits);

				// memory allocation for temporary palette data
				if ((ppal = HeapAlloc(hHeap, 0, sizeof(LOGPALETTE) + wBits * sizeof(PALETTEENTRY))) != NULL)
				{
					ppal->palVersion    = PALVERSION;
					ppal->palNumEntries = wBits;
					GetPaletteEntries(hPalette, 0, wBits, ppal->palPalEntry);
					SetClipboardData(CF_PALETTE, CreatePalette(ppal));
					HeapFree(hHeap, 0, ppal);
				}
			}
			DeleteDC(hBmpDC);
			DeleteObject(hBmp);
			#undef WIDTHBYTES
			#undef PALVERSION
		}
		CloseClipboard();
	}
	return 0;
}

//
// ID_VIEW_RESET
//
static LRESULT OnViewReset(VOID)
{
	if (nState != SM_RUN) return 0;
	if (YesNoMessage(_T("Are you sure you want to press the Reset Button ?"))==IDYES)
	{
		SwitchToState(SM_INVALID);
		CpuReset();							// Chipset setting after power cycle
		SwitchToState(SM_RUN);
	}
	return 0;
}

//
// ID_VIEW_SETTINGS
//
static LRESULT OnViewSettings(VOID)
{
	ReadSettings();		// read from registry

	if (DialogBox(hApp, MAKEINTRESOURCE(IDD_SETTINGS), hWnd, (DLGPROC)SettingsProc) == -1)
		AbortMessage(_T("Settings Dialog Creation Error !"));

	WriteSettings();	// update to registry
	return 0;
}

//
// ID_BACKUP_SAVE
//
static LRESULT OnBackupSave(VOID)
{
	UINT nOldState;
	if (pbyRom == NULL) return 0;
	nOldState = SwitchToState(SM_INVALID);
	SaveBackup();
	SwitchToState(nOldState); 
	return 0;
}

//
// ID_BACKUP_RESTORE
//
static LRESULT OnBackupRestore(VOID)
{
	SwitchToState(SM_INVALID);
	RestoreBackup();
	if (pbyRom) SwitchToState(SM_RUN); 
	return 0;
}

//
// ID_BACKUP_DELETE
//
static LRESULT OnBackupDelete(VOID)
{
	ResetBackup(); 
	return 0;
}


//
// ID_D70X_LOAD
//
static LRESULT OnD70xLoad(BYTE byUnit)		// load an image to a disc unit
{
	SuspendDebugger();						// suspend debugger
	bDbgAutoStateCtrl = FALSE;				// disable automatic debugger state control

	if (nState != SM_RUN)
	{
		InfoMessage(_T("The emulator must be running to load an object."));
		goto cancel;
	}
	if (hp9121_widle())						// wait for idle hp9121 state
	{
		InfoMessage(_T("The HP9121 is busy."));
		goto cancel;
	}
	if (!GetLoadLifFilename())
		goto cancel;
	hp9121_load(byUnit, szBufferFilename);
	UpdateWindowStatus();
cancel:
	bDbgAutoStateCtrl = TRUE;				// enable automatic debugger state control
	ResumeDebugger();
	return 0;
}

//
// ID_D70X_SAVE
//
static LRESULT OnD70xSave(BYTE byUnit)
{
	SuspendDebugger();						// suspend debugger
	bDbgAutoStateCtrl = FALSE;				// disable automatic debugger state control

	if (nState != SM_RUN)
	{
		InfoMessage(_T("The emulator must be running to save a Disk."));
		goto cancel;
	}
	if (hp9121_widle())				// wait for idle hp9121 state
	{
		InfoMessage(_T("The HP9121 is busy."));
		goto cancel;
	}
	_ASSERT(Chipset.D70x_disk[byUnit]!= NULL);
	if (!GetSaveDiskFilename(Chipset.D70x_disk[byUnit]))
		goto cancel;
	hp9121_save(byUnit, szBufferFilename);
cancel:
	bDbgAutoStateCtrl = TRUE;				// enable automatic debugger state control
	ResumeDebugger();
	return 0;
}

//
// ID_D70X_EJECT
//
static LRESULT OnD70xEject(BYTE byUnit)
{
	SuspendDebugger();						// suspend debugger
	bDbgAutoStateCtrl = FALSE;				// disable automatic debugger state control

	if (nState != SM_RUN)
	{
		InfoMessage(_T("The emulator must be running to save an object."));
		goto cancel;
	}
	if (hp9121_widle())				// wait for idle hp9121 state
	{
		InfoMessage(_T("The HP9121 is busy."));
		goto cancel;
	}
	hp9121_eject(byUnit);
	UpdateWindowStatus();
cancel:
	bDbgAutoStateCtrl = TRUE;				// enable automatic debugger state control
	ResumeDebugger();
	return (0);
}

//
// ID_DISK_LOAD
//
static LRESULT OnDiskLoad(VOID)
{
	SuspendDebugger();						// suspend debugger
	bDbgAutoStateCtrl = FALSE;				// disable automatic debugger state control

	if (nState != SM_RUN)
	{
		InfoMessage(_T("The emulator must be running to load an object."));
		goto cancel;
	}
	if (WaitForIdleState())					// wait for idle Basic state
	{
		InfoMessage(_T("The emulator is busy."));
		goto cancel;
	}
	if (!GetLoadLifFilename())
		goto cancel;
	LoadDisk(szBufferFilename);
cancel:
	bDbgAutoStateCtrl = TRUE;				// enable automatic debugger state control
	ResumeDebugger();
	return 0;
}

//
// ID_DISK_SAVE
//
static LRESULT OnDiskSave(VOID)
{
	SuspendDebugger();						// suspend debugger
	bDbgAutoStateCtrl = FALSE;				// disable automatic debugger state control

	if (nState != SM_RUN)
	{
		InfoMessage(_T("The emulator must be running to save an object."));
		goto cancel;
	}
	if (WaitForIdleState())					// wait for idle Basic state
	{
		InfoMessage(_T("The emulator is busy."));
		goto cancel;
	}
	if (!GetSaveLifFilename())
		goto cancel;;
	SaveDisk(szBufferFilename);
cancel:
	bDbgAutoStateCtrl = TRUE;				// enable automatic debugger state control
	ResumeDebugger();
	return 0;
}

//
// ID_LIF_LOAD
//
static LRESULT OnLifLoad(VOID)
{
	SuspendDebugger();						// suspend debugger
	bDbgAutoStateCtrl = FALSE;				// disable automatic debugger state control

	if (nState != SM_RUN)
	{
		InfoMessage(_T("The emulator must be running to load an object."));
		goto cancel;
	}
	if (WaitForIdleState())					// wait for idle Basic state
	{
		InfoMessage(_T("The emulator is busy."));
		goto cancel;
	}
	if (!GetLoadLifFilename())
		goto cancel;
	LoadLif(szBufferFilename);
cancel:
	bDbgAutoStateCtrl = TRUE;				// enable automatic debugger state control
	ResumeDebugger();
	return 0;
}

//
// ID_LIF_SAVE
//
static LRESULT OnLifSave(VOID)		// not really done actually
{
	if (nState != SM_RUN)
	{
		InfoMessage(_T("The emulator must be running to save an object."));
		goto cancel;
	}
	if (WaitForIdleState())					// wait for idle Basic state
	{
		InfoMessage(_T("The emulator is busy."));
		goto cancel;
	}
	if (!GetSaveLifFilename())
		goto cancel;
	SaveLif(szBufferFilename);

cancel:
	bDbgAutoStateCtrl = TRUE;				// enable automatic debugger state control
	ResumeDebugger();
	return 0;
}

static BOOL CALLBACK About(HWND hDlg, UINT message, DWORD wParam, LONG lParam)
{
	switch (message)
	{
	case WM_INITDIALOG:
		SetDlgItemText(hDlg,IDC_VERSION,szNoTitle);
		SetDlgItemText(hDlg,IDC_LICENSE,szLicence);
		return TRUE;
	case WM_COMMAND:
		wParam = LOWORD(wParam);
		if ((wParam==IDOK)||(wParam==IDCANCEL))
		{
			EndDialog(hDlg, wParam);
			return TRUE;
		}
		break;
	}
	return FALSE;
    UNREFERENCED_PARAMETER(lParam);
}

static BOOL CALLBACK Disasm(HWND hDlg, UINT message, DWORD wParam, LONG lParam)
{
	static WORD wAddress;
	LONG  i;
	TCHAR *cpStop, szAddress[256] = _T("0");

	switch (message)
	{
	case WM_INITDIALOG:
		// set fonts & cursor
		SendDlgItemMessage(hDlg,IDC_ADDRESS,WM_SETFONT,(WPARAM)GetStockObject(DEFAULT_GUI_FONT),MAKELPARAM(FALSE,0));
		SendDlgItemMessage(hDlg,IDC_DISASM_ADR,WM_SETFONT,(WPARAM)GetStockObject(DEFAULT_GUI_FONT),MAKELPARAM(FALSE,0));
		SendDlgItemMessage(hDlg,IDC_DISASM_NEXT,WM_SETFONT,(WPARAM)GetStockObject(DEFAULT_GUI_FONT),MAKELPARAM(FALSE,0));
		SendDlgItemMessage(hDlg,IDC_DISASM_COPY,WM_SETFONT,(WPARAM)GetStockObject(DEFAULT_GUI_FONT),MAKELPARAM(FALSE,0));
		SendDlgItemMessage(hDlg,IDCANCEL,WM_SETFONT,(WPARAM)GetStockObject(DEFAULT_GUI_FONT),MAKELPARAM(FALSE,0));
		SetDlgItemText(hDlg,IDC_DISASM_ADR,szAddress);
		init_disassemble(0xFF, 0xFF);
		wAddress = (WORD)(_tcstoul(szAddress,&cpStop,16));	// only 16 bits of address
		return TRUE;
	case WM_COMMAND:
		switch(wParam)
		{
		case IDOK:
			SendDlgItemMessage(hDlg, IDC_DISASM_ADR, EM_SETSEL, 0, -1);
			GetDlgItemText(hDlg, IDC_DISASM_ADR, szAddress, ARRAYSIZEOF(szAddress));
			// test if valid hex address
			for (i = 0; i < (LONG) lstrlen(szAddress); ++i)
			{
				if (_istxdigit(szAddress[i]) == FALSE)
					return FALSE;
			}
			init_disassemble(0xFF, 0xFF);
			wAddress = (WORD)(_tcstoul(szAddress,&cpStop,16));
			// no break
		case IDC_DISASM_NEXT:
			i = wsprintf(szAddress,_T("%04X   "), wAddress);
			wAddress = disassemble(wAddress, &szAddress[i], VIEW_LONG);
			i = SendDlgItemMessage(hDlg, IDC_DISASM_WIN, LB_ADDSTRING, 0, (LPARAM) szAddress);
			SendDlgItemMessage(hDlg, IDC_DISASM_WIN, LB_SELITEMRANGE, FALSE, MAKELPARAM(0,i));
			SendDlgItemMessage(hDlg, IDC_DISASM_WIN, LB_SETSEL, TRUE, i);
			SendDlgItemMessage(hDlg, IDC_DISASM_WIN, LB_SETTOPINDEX, i, 0);
			return TRUE;
		case IDC_DISASM_COPY:
			// copy selected items to clipboard
			CopyItemsToClipboard(GetDlgItem(hDlg, IDC_DISASM_WIN));
			return TRUE;
		case IDCANCEL:
			EndDialog(hDlg, wParam);
			return TRUE;
		}
		break;
	}
	return FALSE;
    UNREFERENCED_PARAMETER(lParam);
}

//
// ID_TOOL_DISASM
//
static LRESULT OnToolDisasm(VOID)			// disasm dialogbox call
{
//	if (pbyRom) SwitchToState(SM_SLEEP);
	if (DialogBox(hApp, MAKEINTRESOURCE(IDD_DISASM), hWnd, (DLGPROC)Disasm) == -1)
		AbortMessage(_T("Disassembler Dialog Box Creation Error !"));
//	if (pbyRom) SwitchToState(SM_RUN);
	return 0;
}

//
// ID_TOOL_DUMPRAM
//
static LRESULT OnToolDumpRam(VOID)			// disasm dialogbox call
{
	HANDLE  hDumpFile = NULL;
	DWORD dwWritten;
	TCHAR  szDumpFilename[MAX_PATH];
	int i;

	i = wsprintf(szDumpFilename,_T("dump_ram_%04d.bin"),nDumpNr);
	szDumpFilename[i++] = 0x00;
	nDumpNr++;

	hDumpFile = CreateFile(szDumpFilename, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, 0, NULL);
	if (hDumpFile == INVALID_HANDLE_VALUE)		// error, couldn't create a new file
	{
		InfoMessage(_T("Can't create the file."));
		return 0;
	}

	if (Chipset.Ram) WriteFile(hDumpFile, Chipset.Ram,  Chipset.RamSize,  &dwWritten, NULL);
	CloseHandle(hDumpFile);

	return 0;
}

//
// ID_ABOUT
//
static LRESULT OnAbout(VOID)
{
	if (DialogBox(hApp, MAKEINTRESOURCE(IDD_ABOUT), hWnd, (DLGPROC)About) == -1)
		AbortMessage(_T("About Dialog Box Creation Error !"));
	return 0;
}

static LRESULT OnLButtonDown(UINT nFlags, WORD x, WORD y)
{
 	if (nState == SM_RUN) MouseButtonDownAt(nFlags, x,y); 
	return 0;
}

static LRESULT OnLButtonUp(UINT nFlags, WORD x, WORD y)
{
 	if (nState == SM_RUN) MouseButtonUpAt(nFlags, x,y); 
	return 0;
}

static LRESULT OnMouseMove(UINT nFlags, WORD x, WORD y)
{
/*	if (nState == SM_RUN) MouseMovesTo(nFlags, x,y); */
	return 0;
}

static LRESULT OnKeyDown(int nVirtKey, DWORD lKeyData)			// for special key down
{
	if (nState == SM_RUN && (lKeyData & 0x40000000) == 0) 
		if (MapVirtualKey(nVirtKey, 2) == 0x00) KeyboardEventDown((BYTE)nVirtKey);
	return 0;
}

static LRESULT OnKeyUp(int nVirtKey, DWORD lKeyData)			// for all key up
{
	if (nState == SM_RUN) KeyboardEventUp((BYTE)nVirtKey); 
	return 0;
	UNREFERENCED_PARAMETER(lKeyData);
}

static LRESULT OnCharDown(TCHAR chCharCode, DWORD lKeyData)		// for ASCII human readable key down
{
	if (nState == SM_RUN)
	if ((lKeyData & 0x40000000) == 0) AsciiEventDown((BYTE)chCharCode);
	return 0;
}

VOID ButtonEvent(BOOL state, UINT nId)
{
	if (state = FALSE)
		return;
	switch (nId)
	{
		case 1:
			OnViewReset();
			break;
		case 2:
		case 3:
		case 4:
		case 5:
#if defined HP87
		case 6:
		case 7:
		case 8:
		case 9:
		case 10:
		case 11:
#endif
			if (Chipset.D70x_disk[nId - 2][0] != 0x00)
				OnD70xEject((BYTE)(nId - 2));
			else
				OnD70xLoad((BYTE)(nId - 2));
			break;
	}
}

LRESULT CALLBACK MainWndProc(HWND hWindow, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_CREATE:          return OnCreate(hWindow);
	case WM_DESTROY:         return OnDestroy(hWindow);
	case WM_PAINT:           return OnPaint(hWindow);
	case WM_ACTIVATE:
		if (LOWORD(wParam)==WA_INACTIVE) break;
	case WM_QUERYNEWPALETTE:
		if (hPalette)
		{
			SelectPalette(hWindowDC, hPalette, FALSE);
			if (RealizePalette(hWindowDC))
			{
				InvalidateRect(hWindow, NULL, TRUE);
				return TRUE;
			}
		}
		return FALSE;
	case WM_PALETTECHANGED:
		if ((HWND)wParam == hWindow) break;
		if (hPalette)
		{
			SelectPalette(hWindowDC, hPalette, FALSE);
			if (RealizePalette(hWindowDC))
			{
				// UpdateColors(hWindowDC);
				InvalidateRect (hWnd, (LPRECT) (NULL), 1);
			}
		}
		return FALSE;
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case ID_FILE_NEW:		return OnFileNew();
		case ID_FILE_OPEN:		return OnFileOpen();
		case ID_FILE_SAVE:		return OnFileSave();
		case ID_FILE_SAVEAS:	return OnFileSaveAs();
		case ID_FILE_CLOSE:		return OnFileClose();
		case ID_FILE_EXIT:		return OnFileExit();
		case ID_COPY_STRING:	return OnCopyString();
		case ID_PASTE_STRING:	return OnPasteString();
		case ID_COPY_SCREEN:	return OnCopyScreen();
		case ID_VIEW_RESET:		return OnViewReset();
		case ID_VIEW_SETTINGS:	return OnViewSettings();
		case ID_BACKUP_SAVE:	return OnBackupSave();
		case ID_BACKUP_RESTORE:	return OnBackupRestore();
		case ID_BACKUP_DELETE:	return OnBackupDelete();
		case ID_LIF_LOAD:		return OnLifLoad();
		case ID_LIF_SAVE:		return OnLifSave();
		case ID_DISK_LOAD:		return OnDiskLoad();
		case ID_DISK_SAVE:		return OnDiskSave();
		case ID_D700_LOAD:		return OnD70xLoad(0);
		case ID_D700_SAVE:		return OnD70xSave(0);
		case ID_D700_EJECT:		return OnD70xEject(0);
		case ID_D701_LOAD:		return OnD70xLoad(1);
		case ID_D701_SAVE:		return OnD70xSave(1);
		case ID_D701_EJECT:		return OnD70xEject(1);
		case ID_D702_LOAD:		return OnD70xLoad(2);
		case ID_D702_SAVE:		return OnD70xSave(2);
		case ID_D702_EJECT:		return OnD70xEject(2);
		case ID_D703_LOAD:		return OnD70xLoad(3);
		case ID_D703_SAVE:		return OnD70xSave(3);
		case ID_D703_EJECT:		return OnD70xEject(3);
#if defined HP87
		case ID_D704_LOAD:		return OnD70xLoad(4);
		case ID_D704_SAVE:		return OnD70xSave(4);
		case ID_D704_EJECT:		return OnD70xEject(4);
		case ID_D705_LOAD:		return OnD70xLoad(5);
		case ID_D705_SAVE:		return OnD70xSave(5);
		case ID_D705_EJECT:		return OnD70xEject(5);
		case ID_D706_LOAD:		return OnD70xLoad(6);
		case ID_D706_SAVE:		return OnD70xSave(6);
		case ID_D706_EJECT:		return OnD70xEject(6);
		case ID_D707_LOAD:		return OnD70xLoad(7);
		case ID_D707_SAVE:		return OnD70xSave(7);
		case ID_D707_EJECT:		return OnD70xEject(7);
		case ID_D708_LOAD:		return OnD70xLoad(8);
		case ID_D708_SAVE:		return OnD70xSave(8);
		case ID_D708_EJECT:		return OnD70xEject(8);
		case ID_D709_LOAD:		return OnD70xLoad(9);
		case ID_D709_SAVE:		return OnD70xSave(9);
		case ID_D709_EJECT:		return OnD70xEject(9);
#endif
		case ID_TOOL_DISASM:	return OnToolDisasm();
		case ID_TOOL_DEBUG:		return OnToolDebug();
		case ID_TOOL_DUMPRAM:	return OnToolDumpRam();
		case ID_ABOUT:			return OnAbout();
		}
		break;
	case WM_SYSCOMMAND:
		switch (LOWORD(wParam))
		{
		case SC_CLOSE: return OnFileExit();
		}
		break;
	case WM_RBUTTONDOWN:
	case WM_LBUTTONDOWN: return OnLButtonDown(wParam, LOWORD(lParam), HIWORD(lParam));
	case WM_LBUTTONUP:   return OnLButtonUp(wParam, LOWORD(lParam), HIWORD(lParam));
//	case WM_MOUSEMOVE:   return OnMouseMove(wParam, LOWORD(lParam), HIWORD(lParam));
	case WM_KEYUP:       return OnKeyUp((int)wParam, lParam);
	case WM_KEYDOWN:     return OnKeyDown((int)wParam, lParam);
	case WM_CHAR:		 return OnCharDown((TCHAR)wParam, lParam);
	}
	return DefWindowProc(hWindow, uMsg, wParam, lParam);
}

int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nCmdShow)
{
	MSG msg;
	WNDCLASS wc;
	RECT rectWindow;
	DWORD dwAffMask;

	hApp = hInst;
	#if defined _UNICODE
	{
		ppArgv = CommandLineToArgvW(GetCommandLine(),&nArgc);
	}
	#else
	{
		nArgc = __argc;						// no. of command line arguments
		ppArgv = (LPCTSTR*) __argv;			// command line arguments
	}
	#endif

	if(!QueryPerformanceFrequency(&lFreq))	// init high resolution counter
	{
		AbortMessage(
			_T("No high resolution timer available.\n")
			_T("This application will now terminate."));
		return FALSE;
	}

	// 16Mo unlimited heap 
	hHeap = HeapCreate(HEAP_NO_SERIALIZE, 16*1024*1024, 0);
	if (hHeap == NULL)					// error, couldn't create a new heap
	{
		AbortMessage(_T("Can't allocate heap."));
		return FALSE;
	}

	// debug file creation
	#if defined DEBUG_ONDISK
	hDebug = CreateFile(_T("debug.txt"), GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, 0, NULL);
	if (hDebug == INVALID_HANDLE_VALUE)		// error, couldn't create a new file
	{
		AbortMessage(_T("Can't create 'debug.txt' file."));
		return FALSE;
	}
	#endif

	wc.style = CS_BYTEALIGNCLIENT;
	wc.lpfnWndProc = (WNDPROC)MainWndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInst;
	wc.hIcon = LoadIcon(hInst, MAKEINTRESOURCE(IDI_HP8X));
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = NULL;
#if defined HP87
	wc.lpszMenuName = MAKEINTRESOURCE(IDR_MENU);
#endif
#if defined HP85
	wc.lpszMenuName = MAKEINTRESOURCE(IDR_MENU85);
#endif
	wc.lpszClassName = _T("CHP8x");

	if (!RegisterClass(&wc))
	{
		AbortMessage(
			_T("CHP8x class registration failed.\n")
			_T("This application will now terminate."));
		return FALSE;
	}

	// Create window
	rectWindow.left   = 0;
	rectWindow.top    = 0;
	rectWindow.right  = 256;
	rectWindow.bottom = 0;
	AdjustWindowRect(&rectWindow, WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX|WS_OVERLAPPED, TRUE);

	hWnd = CreateWindow(_T("CHP8x"), _T("HP8x"),
		WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX|WS_OVERLAPPED,
		CW_USEDEFAULT, CW_USEDEFAULT,
		rectWindow.right  - rectWindow.left,
		rectWindow.bottom - rectWindow.top,
		NULL, NULL, hApp, NULL
		);

	if (hWnd == NULL)
	{
		AbortMessage(_T("Window creation failed."));
		return FALSE;
	}

	GetCurrentDirectory(ARRAYSIZEOF(szCurrentDirectory), szCurrentDirectory);

	ReadSettings();

	UpdateWindowStatus();

	// create auto event handle
	hEventShutdn = CreateEvent(NULL,FALSE,FALSE,NULL);
	if (hEventShutdn == NULL)
	{
		AbortMessage(_T("Event creation failed."));
		DestroyWindow(hWnd);
		return FALSE;
	}

	nState     = SM_RUN;					// init state must be <> nNextState
	nNextState = SM_INVALID;				// go into invalid state
	hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)&WorkerThread, NULL, CREATE_SUSPENDED, &lThreadId);
	if (hThread == NULL)
	{
		CloseHandle(hEventShutdn);			// close event handle
		AbortMessage(_T("Thread creation failed."));
		DestroyWindow(hWnd);
		return FALSE;
	}
	// on multiprocessor machines for QueryPerformanceCounter()
	dwAffMask = SetThreadAffinityMask(hThread,1);
	_ASSERT(dwAffMask != 0);
	ResumeThread(hThread);					// start thread
	while (nState!=nNextState) Sleep(0);	// wait for thread initialized

	_ASSERT(hWnd != NULL);
	_ASSERT(hWindowDC != NULL);

	if (nArgc >= 2)							// use decoded parameter line
	{
		TCHAR szTemp[MAX_PATH+8] = _T("Loading ");
		lstrcat(szTemp, ppArgv[1]);
		SetWindowTitle(szTemp);
		if (OpenDocument(ppArgv[1]))
			goto start;
	}

	ReadLastDocument(szBufferFilename, ARRAYSIZEOF(szBufferFilename));
	if (szBufferFilename[0])
	{
		TCHAR szTemp[MAX_PATH+8] = _T("Loading ");
		lstrcat(szTemp, szBufferFilename);
		SetWindowTitle(szTemp);
		if (OpenDocument(szBufferFilename))
			goto start;
	}

	SetWindowTitle(_T("New Document"));
	if (NewDocument())
	{
		SetWindowTitle(_T("Untitled"));
		goto start;
	}

	ResetDocument();

start:

	ShowWindow(hWnd, nCmdShow);

	if (pbyRom) SwitchToState(SM_RUN);

	while (GetMessage(&msg, NULL, 0, 0))
	{
		if(   (hDlgDebug == NULL || !IsDialogMessage(hDlgDebug, &msg))
		   && (hDlgFind  == NULL || !IsDialogMessage(hDlgFind, &msg)))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	WriteSettings();						// save emulation settings

	IopStop();								// stop peripherals

	CloseHandle(hThread);					// close thread handle
	CloseHandle(hEventShutdn);				// close event handle
	_ASSERT(nState == SM_RETURN);			// emulation thread down?
	ResetDocument();
	ResetBackup();
	_ASSERT(pKml == NULL);					// KML script not closed
	_ASSERT(szTitle == NULL);				// freed allocated memory
	_ASSERT(hPalette == NULL);				// freed resource memory

	CloseWaves();							// free sound memory

#if defined DEBUG_ONDISK
	CloseHandle(hDebug);
#endif
	
	HeapDestroy(hHeap);						// free heap

	return msg.wParam;
	UNREFERENCED_PARAMETER(lpCmdLine);
	UNREFERENCED_PARAMETER(hPrevInst);
}
