/*
 *   Files.c
 *
 *   Copyright 2004 Olivier De Smet
 */

//
//   All stuff for files and documents
//

#include "StdAfx.h"
#include "HP8x.h"
#include "mops.h"
#include "kml.h"
#include "debugger.h"
#include "resource.h"

TCHAR  szHP8xDirectory[MAX_PATH];
TCHAR  szCurrentDirectory[MAX_PATH];
TCHAR  szCurrentKml[MAX_PATH];
TCHAR  szCurrentFilename[MAX_PATH];
TCHAR  szBufferFilename[MAX_PATH];

static TCHAR  szBackupKml[MAX_PATH];
static TCHAR  szBackupFilename[MAX_PATH];

// pointers for roms data
LPBYTE pbyRom = NULL;
LPBYTE pbyRomPage[256] = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 
						  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 
						  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 
						  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 
						  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 
						  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 
						  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 
						  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 
						  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 
						  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 
						  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 
						  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 
						  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 
						  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 
						  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 
						  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};

#if defined HP87
BYTE   cCurrentRomType = 'A';					// Model -> hardware HP87 ("A" by default see KML.c
#endif
#if defined HP85
BYTE   cCurrentRomType = 'B';
#endif

UINT   nCurrentClass = 0;						// Class -> derivate

// document signatures
static BYTE pbySignature[] = "HP8x  Document\xFE";
static HANDLE hCurrentFile = NULL;

static CHIPSET BackupChipset;

BOOL   bBackup = FALSE;

// electronic disk vars
static 	BYTE edFlag = 0x00;
static	LPBYTE edStart = 0x000000;
static	WORD edNsect = 0x0000;

//
// 8160 Kb = 8Mb - 32kb to avoid MSB (used in basic to allocate lines)
// in 16352 Kb config use an 8192 Kb electronic discs to avoid bugs
// p-SYSTEM use up to 6144 Ko as ramdisc
//
static DWORD dwRamSize[] = { 32, 128, 256, 384, 512, 640, 1024, 2048, 4096, 6144, 8160, 16353};
static TCHAR *szDiscType[] = { _T("Mini floppy"), _T("8\" floppy"), _T("Hard disc")};

//################
//#
//#    Low level subroutines
//#
//################

//
// Find the first and only Electronic Disk (can't support multiple e-discs)
//
static BOOL IsEdisk(VOID)
{
	BYTE buff[6];

	edFlag = 0x00;
	edStart = 0x000000;
	edNsect = 0x000;

	if (nState != SM_RUN)
		return FALSE;

#if defined HP87
	Nread_nio(buff, 0x8796, 6);								// buffer
	edFlag = buff[5];										// byte of EDISK
	edStart = (BYTE *)(buff[0]|(buff[1]<<8)|(buff[2]<<16));	// start of EDISK
	edNsect = buff[3]|(buff[4]<<8);							// sectors of EDISK
	if ((edFlag == 0x01) && (edNsect != 0x0000) && (Chipset.RomPage[0xd0]) && (Chipset.RomPage[0xd1]) &&
		(edStart >= (LPBYTE)(0x028000)) && ((edStart + (edNsect << 8)) <= (LPBYTE)(0x008000 + Chipset.RamSize)))
	{
		edStart += Chipset.Ram - (BYTE *)(0x008000);
		return TRUE;
	}
	else
		return FALSE;
#endif
#if defined HP85
	return FALSE;
#endif
}

//
//   Roms emulator crc
//
static WORD CrcRom(VOID)					// calculate fingerprint of ROMs
{
	WORD dwCount;
	WORD dwCrc = 0;
	int i;

	_ASSERT(Chipset.Rom);						// view on ROM
	// use checksum, because it's faster
	for (dwCount = 0; dwCount < 12*1024; dwCount += 1)
		dwCrc += ((WORD *) Chipset.Rom)[dwCount];

	for(i = 0; i < 256; i++)
		if (pbyRomPage[i] != NULL)
			for (dwCount = 0; dwCount < 4*1024; dwCount += 1)
				dwCrc += ((WORD *) pbyRomPage[i])[dwCount];

	return (WORD) dwCrc;
}

//
// calculate a Crc for a 2Ko block and patch it if necessary (for debug only)
//
/* static VOID AssertCrcRom(BYTE rom, LPBYTE addr)
{
	DWORD sum = 0;
	WORD i, crc;

	if (addr != NULL)
	{
		for (i = 0; i < 0xFFF; i++)
			sum += *((WORD *)(addr + (i << 1)));

		crc = (WORD)(sum & 0xFFFF) + (WORD)(sum >> 16);
		crc = ~(crc);

		if (crc != *((WORD *)(addr + 0x1FFE)))
		{
			TCHAR buffer[256];
			int l = wsprintf(buffer,_T("ROM %02X: Crc failed (%04X != %04X), patched\n"), rom, crc, *((WORD *)(addr + 0x1FFE)));

			buffer[l++] = 0x00;
			OutputDebugString(buffer);
			*((WORD *)(addr + 0x1FFE)) = crc;
		}
	}
} */

static LPBYTE  LoadRom(LPCTSTR szRomDirectory, LPCTSTR szFilename)
{
	HANDLE  hRomFile = NULL;
	DWORD dwFileSize, dwRead;
	LPBYTE pRom = NULL;

	SetCurrentDirectory(szHP8xDirectory);
	SetCurrentDirectory(szRomDirectory);
	hRomFile = CreateFile(szFilename,
						  GENERIC_READ,
						  FILE_SHARE_READ,
						  NULL,
						  OPEN_EXISTING,
						  FILE_FLAG_SEQUENTIAL_SCAN,
						  NULL);
	SetCurrentDirectory(szCurrentDirectory);
	if (hRomFile == INVALID_HANDLE_VALUE)
	{
		hRomFile = NULL;
		return NULL;
	}
	dwFileSize = GetFileSize(hRomFile, NULL);
	if ((dwFileSize != 24*1024)  && (dwFileSize != 8*1024))
	{ // file is too small.
		CloseHandle(hRomFile);
		hRomFile = NULL;
		return NULL;
	}

	pRom = HeapAlloc(hHeap, 0, dwFileSize);
	if (pRom == NULL)
	{
		CloseHandle(hRomFile);
		hRomFile = NULL;
		return NULL;
	}

	// load file content
	ReadFile(hRomFile,pRom,dwFileSize,&dwRead,NULL);
	CloseHandle(hRomFile);
	_ASSERT(dwFileSize == dwRead);
	return pRom;
}

static BOOL CALLBACK NewSettingsProc(HWND hDlg, UINT message, DWORD wParam, LONG lParam)
{
	HWND hList;
	UINT nIndex, i, j, k;
	TCHAR buffer[32];

	switch (message)
	{
	case WM_INITDIALOG:
		// init rom checkbox
#if defined HP87
		if (Chipset.RomPage[0x0E])
			CheckDlgButton(hDlg, IDC_SETROM0E , TRUE);
		else
			CheckDlgButton(hDlg, IDC_SETROM0E , FALSE);
		if (Chipset.RomPage[0x18])
			CheckDlgButton(hDlg, IDC_SETROM18 , TRUE);
		else
			CheckDlgButton(hDlg, IDC_SETROM18 , FALSE);
		if (Chipset.RomPage[0x28])
			CheckDlgButton(hDlg, IDC_SETROM28 , TRUE);
		else
			CheckDlgButton(hDlg, IDC_SETROM28 , FALSE);
		if (Chipset.RomPage[0x38])
			CheckDlgButton(hDlg, IDC_SETROM38 , TRUE);
		else
			CheckDlgButton(hDlg, IDC_SETROM38 , FALSE);
		if (Chipset.RomPage[0xB0])
			CheckDlgButton(hDlg, IDC_SETROMB0 , TRUE);
		else
			CheckDlgButton(hDlg, IDC_SETROMB0 , FALSE);
		if (Chipset.RomPage[0xC0])
			CheckDlgButton(hDlg, IDC_SETROMC0 , TRUE);
		else
			CheckDlgButton(hDlg, IDC_SETROMC0 , FALSE);
		if (Chipset.RomPage[0xCF])
			CheckDlgButton(hDlg, IDC_SETROMCF , TRUE);
		else
			CheckDlgButton(hDlg, IDC_SETROMCF , FALSE);
		if (Chipset.RomPage[0xD0])
			CheckDlgButton(hDlg, IDC_SETROMD0 , TRUE);
		else
			CheckDlgButton(hDlg, IDC_SETROMD0 , FALSE);
		if (Chipset.RomPage[0xD1])
			CheckDlgButton(hDlg, IDC_SETROMD1 , TRUE);
		else
			CheckDlgButton(hDlg, IDC_SETROMD1 , FALSE);
		if (Chipset.RomPage[0xE7])
			CheckDlgButton(hDlg, IDC_SETROME7 , TRUE);
		else
			CheckDlgButton(hDlg, IDC_SETROME7 , FALSE);
		if (Chipset.RomPage[0xF0])
			CheckDlgButton(hDlg, IDC_SETROMF0 , TRUE);
		else
			CheckDlgButton(hDlg, IDC_SETROMF0 , FALSE);
#endif
#if defined HP85
		if (Chipset.RomPage[0x08])
			CheckDlgButton(hDlg, IDC_SETROM08 , TRUE);
		else
			CheckDlgButton(hDlg, IDC_SETROM08 , FALSE);
		if (Chipset.RomPage[0x28])
			CheckDlgButton(hDlg, IDC_SETROM28 , TRUE);
		else
			CheckDlgButton(hDlg, IDC_SETROM28 , FALSE);
		if (Chipset.RomPage[0xB0])
			CheckDlgButton(hDlg, IDC_SETROMB0 , TRUE);
		else
			CheckDlgButton(hDlg, IDC_SETROMB0 , FALSE);
		if (Chipset.RomPage[0xC0])
			CheckDlgButton(hDlg, IDC_SETROMC0 , TRUE);
		else
			CheckDlgButton(hDlg, IDC_SETROMC0 , FALSE);
		if (Chipset.RomPage[0xCF])
			CheckDlgButton(hDlg, IDC_SETROMCF , TRUE);
		else
			CheckDlgButton(hDlg, IDC_SETROMCF , FALSE);
		if (Chipset.RomPage[0xD0])
			CheckDlgButton(hDlg, IDC_SETROMD0 , TRUE);
		else
			CheckDlgButton(hDlg, IDC_SETROMD0 , FALSE);
		if (Chipset.RomPage[0xD1])
			CheckDlgButton(hDlg, IDC_SETROMD1 , TRUE);
		else
			CheckDlgButton(hDlg, IDC_SETROMD1 , FALSE);
		if (Chipset.RomPage[0xE0])
			CheckDlgButton(hDlg, IDC_SETROME0 , TRUE);
		else
			CheckDlgButton(hDlg, IDC_SETROME0 , FALSE);
		if (Chipset.RomPage[0xE8])
			CheckDlgButton(hDlg, IDC_SETROME8 , TRUE);
		else
			CheckDlgButton(hDlg, IDC_SETROME8 , FALSE);
		if (Chipset.RomPage[0xF0])
			CheckDlgButton(hDlg, IDC_SETROMF0 , TRUE);
		else
			CheckDlgButton(hDlg, IDC_SETROMF0 , FALSE);
#endif

		if (Chipset.iop[7].iop)
			CheckDlgButton(hDlg, IDC_SETSC07 , TRUE);
		else
			CheckDlgButton(hDlg, IDC_SETSC07 , FALSE);
		if (Chipset.iop[10].iop)
			CheckDlgButton(hDlg, IDC_SETSC10 , TRUE);
		else
			CheckDlgButton(hDlg, IDC_SETSC10 , FALSE);
		// init ram list
		hList = GetDlgItem(hDlg,IDC_SETRAM);
		SendMessage(hList, CB_RESETCONTENT, 0, 0);
		for (i = 0, k = 0; i < 11; i++)
		{
			j = wsprintf(buffer, "%d Kb", dwRamSize[i]);
			buffer[j] = 0x00;
			nIndex = SendMessage(hList, CB_ADDSTRING, 0, (LPARAM)buffer);
			SendMessage(hList, CB_SETITEMDATA, nIndex, i);
			if (Chipset.RamSize == _KB(dwRamSize[i]))
				k = i;
		}
		SendMessage(hList, CB_SETCURSEL, k, 0);
		// init disc list
		hList = GetDlgItem(hDlg,IDC_SETDISC70);
		SendMessage(hList, CB_RESETCONTENT, 0, 0);
		for (i = 0; i < 3; i++)
		{
			nIndex = SendMessage(hList, CB_ADDSTRING, 0, (LPARAM)szDiscType[i]);
			SendMessage(hList, CB_SETITEMDATA, nIndex, i);
		}
		SendMessage(hList, CB_SETCURSEL, Chipset.D70x_type, 0);

		return TRUE;
	case WM_COMMAND:
		if (wParam == IDOK)
		{
			if (szCurrentKml[0])
			{
				ResetDocument();

				hList = GetDlgItem(hDlg,IDC_SETRAM);
				nIndex = SendMessage(hList, CB_GETCURSEL, 0, 0);
				Chipset.RamSize = _KB(dwRamSize[nIndex]);		// standard RAM size + EMSsize in KB 

				hList = GetDlgItem(hDlg,IDC_SETDISC70);
				nIndex = SendMessage(hList, CB_GETCURSEL, 0, 0);
				Chipset.D70x_type = (BYTE)(nIndex);				// type of disk

				// get roms checkbox values
#if defined HP87
				Chipset.RomPage[0x00] = 1;
				Chipset.RomPage[0x01] = 1;
				Chipset.RomPage[0x0E] = IsDlgButtonChecked(hDlg,IDC_SETROM0E) ? 1 : 0;
				Chipset.RomPage[0x18] = IsDlgButtonChecked(hDlg,IDC_SETROM18) ? 1 : 0;
				Chipset.RomPage[0x28] = IsDlgButtonChecked(hDlg,IDC_SETROM28) ? 1 : 0;
				Chipset.RomPage[0x38] = IsDlgButtonChecked(hDlg,IDC_SETROM38) ? 1 : 0;
				Chipset.RomPage[0xB0] = IsDlgButtonChecked(hDlg,IDC_SETROMB0) ? 1 : 0;
				Chipset.RomPage[0xB1] = IsDlgButtonChecked(hDlg,IDC_SETROMB0) ? 1 : 0;
				Chipset.RomPage[0xC0] = IsDlgButtonChecked(hDlg,IDC_SETROMC0) ? 1 : 0;
				Chipset.RomPage[0xCF] = IsDlgButtonChecked(hDlg,IDC_SETROMCF) ? 1 : 0;
				Chipset.RomPage[0xD0] = IsDlgButtonChecked(hDlg,IDC_SETROMD0) ? 1 : 0;
				Chipset.RomPage[0xD1] = IsDlgButtonChecked(hDlg,IDC_SETROMD1) ? 1 : 0;
				Chipset.RomPage[0xE7] = IsDlgButtonChecked(hDlg,IDC_SETROME7) ? 1 : 0;
				Chipset.RomPage[0xE8] = IsDlgButtonChecked(hDlg,IDC_SETROME7) ? 1 : 0;
				Chipset.RomPage[0xF0] = IsDlgButtonChecked(hDlg,IDC_SETROMF0) ? 1 : 0;
#endif
#if defined HP85
				Chipset.RomPage[0x00] = 1;
				Chipset.RomPage[0x08] = IsDlgButtonChecked(hDlg,IDC_SETROM08) ? 1 : 0;	// 9915 development ...
				Chipset.RomPage[0x28] = IsDlgButtonChecked(hDlg,IDC_SETROM28) ? 1 : 0;
				Chipset.RomPage[0xB0] = IsDlgButtonChecked(hDlg,IDC_SETROMB0) ? 1 : 0;
				Chipset.RomPage[0xC0] = IsDlgButtonChecked(hDlg,IDC_SETROMC0) ? 1 : 0;
				Chipset.RomPage[0xCF] = IsDlgButtonChecked(hDlg,IDC_SETROMCF) ? 1 : 0;
				Chipset.RomPage[0xD0] = IsDlgButtonChecked(hDlg,IDC_SETROMD0) ? 1 : 0;
				Chipset.RomPage[0xD1] = IsDlgButtonChecked(hDlg,IDC_SETROMD1) ? 1 : 0;
				Chipset.RomPage[0xE0] = IsDlgButtonChecked(hDlg,IDC_SETROME0) ? 1 : 0;	// services
				Chipset.RomPage[0xE8] = IsDlgButtonChecked(hDlg,IDC_SETROME8) ? 1 : 0;
				Chipset.RomPage[0xF0] = IsDlgButtonChecked(hDlg,IDC_SETROMF0) ? 1 : 0;

#endif

				Chipset.iop[7].iop = IsDlgButtonChecked(hDlg,IDC_SETSC07) ? 1 : 0;
				Chipset.iop[10].iop = IsDlgButtonChecked(hDlg,IDC_SETSC10) ? 2 : 0;
				EndDialog(hDlg, wParam);
			}
		}
		if (wParam == IDCANCEL)
		{
			EndDialog(hDlg, wParam);
		}
		if (wParam == IDC_SETKML)
		{
			DisplayChooseKml(0);
		}
		break;
	} 
	return FALSE;
    UNREFERENCED_PARAMETER(lParam);
}

//################
//#
//#    Public functions
//#
//################

//
//  Window Position Tools
//
VOID SetWindowLocation(HWND hWnd,INT nPosX,INT nPosY)
{
	WINDOWPLACEMENT wndpl;
	RECT *pRc = &wndpl.rcNormalPosition;

	wndpl.length = sizeof(wndpl);
	GetWindowPlacement(hWnd,&wndpl);
	pRc->right = pRc->right - pRc->left + nPosX;
	pRc->bottom = pRc->bottom - pRc->top + nPosY;
	pRc->left = nPosX;
	pRc->top = nPosY;
	SetWindowPlacement(hWnd,&wndpl);
}

//
//   Roms
//

BOOL PatchRom(LPCTSTR szFilename)
{
	return TRUE;
}

BOOL MapRom(LPCTSTR szRomDirectory)
{
	INT   i;

	if (pbyRom != NULL)
	{
		return FALSE;
	}

#if defined HP87

#if defined VRAM64K
	Chipset.Rom = pbyRom = LoadRom(szRomDirectory,"87_sysp.bin");
//	AssertCrcRom(0, pbyRom);			// crc patch done
#else
	Chipset.Rom = pbyRom = LoadRom(szRomDirectory,"87_sys.bin");
#endif

	for(i = 0; i< 256; i++)
		pbyRomPage[i] = NULL;
	pbyRomPage[0] = LoadRom(szRomDirectory, "87_00.bin");				// system rom 000						hp87
#if defined VRAM64K
	pbyRomPage[1] = LoadRom(szRomDirectory, "87_01p.bin");				// system rom 001						hp87
//	AssertCrcRom(1, pbyRomPage[1]);		// crc patch done
#else
	pbyRomPage[1] = LoadRom(szRomDirectory, "87_01.bin");				// system rom 001						hp87
#endif
	pbyRomPage[14] = LoadRom(szRomDirectory, "87_0e.bin");			// paged  rom 016 MIKSAM
	pbyRomPage[24] = LoadRom(szRomDirectory, "87_18.bin");			// paged  rom 030 LANGUAGE				hp86b german
	pbyRomPage[40] = LoadRom(szRomDirectory, "87_28.bin");			// paged  rom 050 ASSEMBLER
	pbyRomPage[56] = LoadRom(szRomDirectory, "87_38.bin");			// paged  rom 070 A Koppel
	pbyRomPage[176] = LoadRom(szRomDirectory, "87_b0.bin");			// paged  rom 260 MATRIX1
	pbyRomPage[177] = LoadRom(szRomDirectory, "87_b1.bin");			// paged  rom 261 MATRIX2
	pbyRomPage[192] = LoadRom(szRomDirectory, "87_c0.bin");			// paged  rom 300 IO
	pbyRomPage[207] = LoadRom(szRomDirectory, "87_cf.bin");			// paged  rom 317 EMS
#if defined VRAM64K
	pbyRomPage[208] = LoadRom(szRomDirectory, "87_d0p.bin");		// paged  rom 320 MASS STORAGE			hp87
//	AssertCrcRom(208, pbyRomPage[208]);		// crc patch done
#else
	pbyRomPage[208] = LoadRom(szRomDirectory, "87_d0.bin");			// paged  rom 320 MASS STORAGE			hp87
#endif
	pbyRomPage[209] = LoadRom(szRomDirectory, "87_d1.bin");			// paged  rom 321 ELECTRONIC DISK		hp86b
	pbyRomPage[231] = LoadRom(szRomDirectory, "87_e7.bin");			// paged  rom 347 ADV-PROG-2-RevB
	pbyRomPage[232] = LoadRom(szRomDirectory, "87_e8.bin");			// paged  rom 350 ADV-PROG-1-RevB
	pbyRomPage[240] = LoadRom(szRomDirectory, "87_f0.bin");			// paged  rom 360 PRINTER PLOTTER

#endif
#if defined HP85

	Chipset.Rom = pbyRom = LoadRom(szRomDirectory,"85b_sys.bin");

	for(i = 0; i< 256; i++)
		pbyRomPage[i] = NULL;
	pbyRomPage[0] = LoadRom(szRomDirectory, "85b_00.bin");			// system rom 000						hp85
	pbyRomPage[8] = LoadRom(szRomDirectory, "85_08.bin");			// system rom 010 ???
	pbyRomPage[40] = LoadRom(szRomDirectory, "85_28.bin");			// paged  rom 050 ASSEMBLER
	pbyRomPage[176] = LoadRom(szRomDirectory, "85_b0.bin");			// paged  rom 260 MATRIX
	pbyRomPage[192] = LoadRom(szRomDirectory, "85b_c0.bin");		// paged  rom 300 IO
	pbyRomPage[207] = LoadRom(szRomDirectory, "85_cf.bin");			// paged  rom 317 EMS
	pbyRomPage[208] = LoadRom(szRomDirectory, "85b_d0.bin");		// paged  rom 320 MASS STORAGE			hp85
	pbyRomPage[209] = LoadRom(szRomDirectory, "85b_d1.bin");		// paged  rom 321 ELECTRONIC DISK		hp85b
	pbyRomPage[224] = LoadRom(szRomDirectory, "85_e0.bin");			// paged  rom 340 SERVICE
	pbyRomPage[232] = LoadRom(szRomDirectory, "85_e8.bin");			// paged  rom 350 ADV-PROG-RevB
	pbyRomPage[240] = LoadRom(szRomDirectory, "85_f0.bin");			// paged  rom 360 PRINTER PLOTTER

#endif

	return TRUE;
}

VOID UnmapRom(VOID)
{
	WORD i;

	if (pbyRom==NULL) return;
	HeapFree(hHeap, 0, pbyRom);
	pbyRom = NULL;
	for (i = 0; i < 256; i++)
		if (pbyRomPage[i])
		{
			HeapFree(hHeap, 0, pbyRomPage[i]);
			pbyRomPage[i] = NULL;
		}
}

//
//   Documents
//

VOID ResetDocument(VOID)
{
	if (szCurrentKml[0])
	{
		KillKML();
	}
	if (hCurrentFile)
	{
		CloseHandle(hCurrentFile);
		hCurrentFile = NULL;
	}
//	szCurrentKml[0] = 0;					// preserve the current KML
	szCurrentFilename[0]=0;
	if (Chipset.Ram) 
	{
		HeapFree(hHeap, 0, Chipset.Ram);
		Chipset.Ram = NULL;
	}
	if (Chipset.Video) 
	{
		HeapFree(hHeap, 0, Chipset.Video);
		Chipset.Video = NULL;
	}
//	ZeroMemory(&Chipset,sizeof(Chipset));

#if defined HP87

#if defined VRAM64K
	Chipset.VideoSize = _KB(64);			// video Ram size in KB
#else
	Chipset.VideoSize = _KB(16);			// video Ram size in KB
#endif

#endif
#if defined HP85

	Chipset.VideoSize = _KB(8);				// video Ram size in KB

#endif

	UpdateWindowStatus();
}

BOOL NewDocument(VOID)
{
#if defined HP87
	if (DialogBox(hApp, MAKEINTRESOURCE(IDD_NEWSETTINGS), hWnd, (DLGPROC)NewSettingsProc) == -1)
		AbortMessage(_T("New Emulation Dialog Creation Error !"));
#endif
#if defined HP85
	if (DialogBox(hApp, MAKEINTRESOURCE(IDD_NEWSETTINGS85), hWnd, (DLGPROC)NewSettingsProc) == -1)
		AbortMessage(_T("New Emulation Dialog Creation Error !"));
#endif
	if (!InitKML(szCurrentKml,FALSE)) goto restore;

#if defined HP87

#if defined VRAM64K
	if (nLcdxZoom != 3)
	{
		AbortMessage(_T("Hp86b+ only with a Zoom 3 KML."));
		goto restore;
	}
#endif

#endif

	Chipset.type = cCurrentRomType;

	// allocate memory
	if (Chipset.Ram == NULL)
	{
		Chipset.Ram = (LPBYTE)HeapAlloc(hHeap, 0, Chipset.RamSize);
		_ASSERT(Chipset.Ram != NULL);
	}
	if (Chipset.Video == NULL)
	{
		Chipset.Video = (LPBYTE)HeapAlloc(hHeap, 0, Chipset.VideoSize);
		_ASSERT(Chipset.Video != NULL);
	}

	CpuReset();
	LoadBreakpointList(NULL);				// clear debugger breakpoint list
	SaveBackup();
	return TRUE;
 restore:
	RestoreBackup();
	ResetBackup();
	SetWindowLocation(hWnd,Chipset.nPosX,Chipset.nPosY);

#if defined HP85
	SetWindowLocation(hWndPrt85,Chipset.nPPosX,Chipset.nPPosY);
#endif

	return FALSE;
}

BOOL OpenDocument(LPCTSTR szFilename)
{
	HANDLE  hFile = INVALID_HANDLE_VALUE;
	DWORD   lBytesRead,lSizeofChipset;
	BYTE    pbyFileSignature[16];
	UINT    ctBytesCompared;
	UINT    nLength;

	SaveBackup();
	ResetDocument();

	// Open file
	if (lstrcmpi(szBackupFilename, szFilename)==0)
	{
		if (YesNoMessage(_T("Do you want to reload this document ?")) == IDNO)
			goto restore;
	}
	hFile = CreateFile(szFilename, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
	{
		AbortMessage(_T("This file is missing or already loaded in another instance of Hp8x."));
		goto restore;
	}

	// Read and Compare signature
	ReadFile(hFile, pbyFileSignature, sizeof(pbyFileSignature), &lBytesRead, NULL);
	for (ctBytesCompared=0; ctBytesCompared<sizeof(pbyFileSignature); ctBytesCompared++)
	{
		if (pbyFileSignature[ctBytesCompared]!=pbySignature[ctBytesCompared])
		{
			AbortMessage(_T("This file is not a valid Hp8x document."));
			goto restore;
		}
	}

	switch (pbyFileSignature[14])
	{
	case 0xFE: // Hp8x 1.0 format
		ReadFile(hFile,&nLength,sizeof(nLength),&lBytesRead,NULL);
		#if defined _UNICODE
		{
			LPSTR szTmp = HeapAlloc(hHeap, 0, nLength);
			if (szTmp == NULL)
			{
				AbortMessage(_T("Memory Allocation Failure."));
				goto restore;
			}
			ReadFile(hFile, szTmp, nLength, &lBytesRead, NULL);
			MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, szTmp, lBytesRead,
								szCurrentKml, ARRAYSIZEOF(szCurrentKml));
			HeapFree(hHeap, 0, szTmp);
		}
		#else
		{
			ReadFile(hFile, szCurrentKml, nLength, &lBytesRead, NULL);
		}
		#endif
		if (nLength != lBytesRead) goto read_err;
		szCurrentKml[nLength] = 0;
		break;
	default:
		AbortMessage(_T("This file is for an unknown version of Hp8x."));
		goto restore;
	}

	// read chipset size inside file
	ReadFile(hFile, &lSizeofChipset, sizeof(lSizeofChipset), &lBytesRead, NULL);
	if (lBytesRead != sizeof(lSizeofChipset)) goto read_err;
	if (lSizeofChipset <= sizeof(Chipset))	// actual or older chipset version
	{
		// read chipset content
		ZeroMemory(&Chipset,sizeof(Chipset)); // init chipset
		ReadFile(hFile, &Chipset, lSizeofChipset, &lBytesRead, NULL);
	}
	else									// newer chipset version
	{
		// read my used chipset content
		ReadFile(hFile, &Chipset, sizeof(Chipset), &lBytesRead, NULL);

		// skip rest of chipset
		SetFilePointer(hFile, lSizeofChipset-sizeof(Chipset), NULL, FILE_CURRENT);
		lSizeofChipset = sizeof(Chipset);
	}
	if (lBytesRead != lSizeofChipset) goto read_err;

#if defined HP87

#if defined VRAM64K
	if (Chipset.VideoSize != _KB(64))
#else
	if (Chipset.VideoSize != _KB(16))
#endif

#endif
#if defined HP85

	if (Chipset.VideoSize != _KB(8))

#endif

	{
		AbortMessage(_T("This file is for a different VRAM model of Hp8x."));
		goto restore;
	}

	SetWindowLocation(hWnd,Chipset.nPosX,Chipset.nPosY);

#if defined HP85
	SetWindowLocation(hWndPrt85,Chipset.nPPosX,Chipset.nPPosY);
#endif

	if (szCurrentKml == NULL)
	{
		if (!DisplayChooseKml(Chipset.type))
			goto restore;
	}
	while (TRUE)
	{
		BOOL bOK;

		bOK = InitKML(szCurrentKml,FALSE);

#if defined HP87

#if defined VRAM64K
		if (nLcdxZoom != 3)
		{
			AbortMessage(_T("Hp86b+ only with a Zoom 3 KML."));
			bOK = FALSE;
		}
#endif

#endif

		bOK = bOK && (cCurrentRomType == Chipset.type);
		if (bOK) break;

		KillKML();
		if (!DisplayChooseKml(Chipset.type))
			goto restore;
	}

	if (Chipset.Ram)					// RAM
	{
		Chipset.Ram = (LPBYTE)HeapAlloc(hHeap, 0, Chipset.RamSize);
		if (Chipset.Ram == NULL)
		{
			AbortMessage(_T("Memory Allocation Failure."));
			goto restore;
		}

		ReadFile(hFile, Chipset.Ram, Chipset.RamSize, &lBytesRead, NULL);
		if (lBytesRead != Chipset.RamSize) goto read_err;
	}
	if (Chipset.Video)					// Video
	{
		Chipset.Video = (LPBYTE)HeapAlloc(hHeap, 0, Chipset.VideoSize);
		if (Chipset.Video == NULL)
		{
			AbortMessage(_T("Memory Allocation Failure."));
			goto restore;
		}

		ReadFile(hFile, Chipset.Video, Chipset.VideoSize, &lBytesRead, NULL);
		if (lBytesRead != Chipset.VideoSize) goto read_err;
	}

	IopReload();							// reload functions for IOP's

	LoadBreakpointList(hFile);				// load debugger breakpoint list

	if (Chipset.RomCrc != CrcRom())		// ROM changed
	{
		CpuReset();
	}

	lstrcpy(szCurrentFilename, szFilename);
	_ASSERT(hCurrentFile == NULL);
	hCurrentFile = hFile;
	SetWindowTitle(szCurrentFilename);
	UpdateWindowStatus();
	UpdateAnnunciators();
	return TRUE;

 read_err:
	AbortMessage(_T("This file must be truncated, and cannot be loaded."));
 restore:
	if (INVALID_HANDLE_VALUE != hFile)		// close if valid handle
		CloseHandle(hFile);
	RestoreBackup();
	ResetBackup();
	return FALSE;
}

BOOL SaveDocument(VOID)
{
	DWORD           lBytesWritten;
	DWORD           lSizeofChipset;
	UINT            nLength;
	WINDOWPLACEMENT wndpl;

	if (hCurrentFile == NULL) return FALSE;

	_ASSERT(hWnd);							// window open
	wndpl.length = sizeof(wndpl);			// update saved window position
	GetWindowPlacement(hWnd, &wndpl);
	Chipset.nPosX = (SWORD) wndpl.rcNormalPosition.left;
	Chipset.nPosY = (SWORD) wndpl.rcNormalPosition.top;

#if defined HP85
	wndpl.length = sizeof(wndpl);			// update saved window position
	GetWindowPlacement(hWndPrt85, &wndpl);
	Chipset.nPPosX = (SWORD) wndpl.rcNormalPosition.left;
	Chipset.nPPosY = (SWORD) wndpl.rcNormalPosition.top;
#endif

	SetFilePointer(hCurrentFile,0,NULL,FILE_BEGIN);

	if (!WriteFile(hCurrentFile, pbySignature, sizeof(pbySignature), &lBytesWritten, NULL))
	{
		AbortMessage(_T("Could not write into file !"));
		return FALSE;
	}

	Chipset.RomCrc = CrcRom();				// save fingerprint of ROM

	nLength = lstrlen(szCurrentKml);
	WriteFile(hCurrentFile, &nLength, sizeof(nLength), &lBytesWritten, NULL);
	#if defined _UNICODE
	{
		LPSTR szTmp = HeapAlloc(hHeap, 0, nLength);
		if (szTmp != NULL)
		{
			WideCharToMultiByte(CP_ACP, WC_COMPOSITECHECK,
								szCurrentKml, nLength,
								szTmp, nLength, NULL, NULL);
			WriteFile(hCurrentFile, szTmp, nLength, &lBytesWritten, NULL);
			HeapFree(hHeap, 0, szTmp);
		}
	}
	#else
	{
		WriteFile(hCurrentFile, szCurrentKml, nLength, &lBytesWritten, NULL);
	}
	#endif
	lSizeofChipset = sizeof(Chipset);
	WriteFile(hCurrentFile, &lSizeofChipset, sizeof(lSizeofChipset), &lBytesWritten, NULL);
	WriteFile(hCurrentFile, &Chipset, lSizeofChipset, &lBytesWritten, NULL);
	if (Chipset.Ram) WriteFile(hCurrentFile, Chipset.Ram,  Chipset.RamSize,  &lBytesWritten, NULL);
	if (Chipset.Video) WriteFile(hCurrentFile, Chipset.Video,  Chipset.VideoSize,  &lBytesWritten, NULL);
	SaveBreakpointList(hCurrentFile);		// save debugger breakpoint list
	SetEndOfFile(hCurrentFile);				// cut the rest 
	return TRUE;
}

BOOL SaveDocumentAs(LPCTSTR szFilename)
{
	HANDLE hFile;

	if (hCurrentFile)						// already file in use
	{
		CloseHandle(hCurrentFile);			// close it, even it's same, so data always will be saved
		hCurrentFile = NULL;
	}
	hFile = CreateFile(szFilename, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, 0, NULL);
	if (hFile == INVALID_HANDLE_VALUE)		// error, couldn't create a new file
	{
		AbortMessage(_T("This file must be currently used by another instance of Hp8x."));
		return FALSE;
	}
	lstrcpy(szCurrentFilename, szFilename);	// save new file name
	hCurrentFile = hFile;					// and the corresponding handle
	SetWindowTitle(szCurrentFilename);		// update window title line
	UpdateWindowStatus();					// and draw it 
	return SaveDocument();					// save current content
}

//
//   Backups
//

BOOL SaveBackup(VOID)
{
	WINDOWPLACEMENT wndpl;

	if (pbyRom == NULL) return FALSE;

	wndpl.length = sizeof(wndpl);			// update saved window position
	GetWindowPlacement(hWnd, &wndpl);
	Chipset.nPosX = (SWORD) wndpl.rcNormalPosition.left;
	Chipset.nPosY = (SWORD) wndpl.rcNormalPosition.top;

#if defined HP85
	wndpl.length = sizeof(wndpl);			// update saved window position
	GetWindowPlacement(hWndPrt85, &wndpl);
	Chipset.nPPosX = (SWORD) wndpl.rcNormalPosition.left;
	Chipset.nPPosY = (SWORD) wndpl.rcNormalPosition.top;
#endif

	lstrcpy(szBackupFilename, szCurrentFilename);
	lstrcpy(szBackupKml, szCurrentKml);
	CopyMemory(&BackupChipset, &Chipset, sizeof(Chipset));
	if (Chipset.Ram)						// standard RAM
	{
		BackupChipset.Ram = (LPBYTE)HeapAlloc(hHeap, 0, Chipset.RamSize);
		CopyMemory(BackupChipset.Ram, Chipset.Ram, Chipset.RamSize);
	}
	if (Chipset.Video)						// video RAM
	{
		BackupChipset.Video = (LPBYTE)HeapAlloc(hHeap, 0, Chipset.VideoSize);
		CopyMemory(BackupChipset.Video, Chipset.Video, Chipset.VideoSize);
	}
	bBackup = TRUE;
	UpdateWindowStatus();
	return TRUE;
}

BOOL RestoreBackup(VOID)
{
	if (!bBackup) return FALSE;
	ResetDocument();
	if (!InitKML(szBackupKml,TRUE))
	{
		InitKML(szCurrentKml,TRUE);
		return FALSE;
	}
	lstrcpy(szCurrentKml, szBackupKml);
	lstrcpy(szCurrentFilename, szBackupFilename);
	if (szCurrentFilename[0])
	{
		hCurrentFile = CreateFile(szCurrentFilename, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
		if (hCurrentFile == INVALID_HANDLE_VALUE)
		{
			hCurrentFile = NULL;
			szCurrentFilename[0] = 0;
		}
	}
	CopyMemory(&Chipset, &BackupChipset, sizeof(Chipset));
	if (BackupChipset.Ram)					// standard RAM
	{
		Chipset.Ram = (LPBYTE)HeapAlloc(hHeap, 0, Chipset.RamSize);
		CopyMemory(Chipset.Ram, BackupChipset.Ram, Chipset.RamSize);
	}
	if (BackupChipset.Video)				// video RAM
	{
		Chipset.Video = (LPBYTE)HeapAlloc(hHeap, 0, Chipset.VideoSize);
		CopyMemory(Chipset.Video, BackupChipset.Video, Chipset.VideoSize);
	}
	SetWindowLocation(hWnd,Chipset.nPosX,Chipset.nPosY);
	UpdateWindowStatus();

#if defined HP85
	SetWindowLocation(hWndPrt85,Chipset.nPPosX,Chipset.nPPosY);
#endif

	return TRUE;
}

BOOL ResetBackup(VOID)
{
	if (!bBackup) return FALSE;
	szBackupFilename[0] = 0;
	szBackupKml[0] = 0;
	if (BackupChipset.Ram) HeapFree(hHeap, 0, BackupChipset.Ram);
	if (BackupChipset.Video) HeapFree(hHeap, 0, BackupChipset.Video);
	ZeroMemory(&BackupChipset,sizeof(BackupChipset));
	bBackup = FALSE;
	UpdateWindowStatus();
	return TRUE;
}

//
//   Open File Common Dialog Boxes
//

static VOID InitializeOFN(LPOPENFILENAME ofn)
{
	ZeroMemory((LPVOID)ofn, sizeof(OPENFILENAME));
	ofn->lStructSize = sizeof(OPENFILENAME);
	ofn->hwndOwner = hWnd;
	ofn->Flags = OFN_EXPLORER|OFN_HIDEREADONLY;
}

BOOL GetOpenFilename(VOID)
{
	OPENFILENAME ofn;

	InitializeOFN(&ofn);
	ofn.lpstrFilter =
		_T("Hp8x Document (*.E8x)\0*.E8x\0")
		_T("\0\0");
	ofn.lpstrDefExt = _T("E8x");			// HP8x
	ofn.nFilterIndex = 1;
	ofn.lpstrFile = HeapAlloc(hHeap, 0, 512);
	ofn.lpstrFile[0] = 0;
	ofn.nMaxFile = 512;
	ofn.Flags |= OFN_FILEMUSTEXIST|OFN_PATHMUSTEXIST;
	if (GetOpenFileName(&ofn) == FALSE)
	{
		HeapFree(hHeap, 0, ofn.lpstrFile);
		return FALSE;
	}
	lstrcpy(szBufferFilename, ofn.lpstrFile);
	HeapFree(hHeap, 0, ofn.lpstrFile); 
	return TRUE;
}

BOOL GetSaveAsFilename(VOID)
{
 	OPENFILENAME ofn;

	InitializeOFN(&ofn);
	ofn.lpstrFilter =
		_T("Hp8x Document (*.E8x)\0*.E8x\0")
		_T("\0\0");
	ofn.lpstrDefExt = _T("E8x");			// HP8x
	ofn.nFilterIndex = 1;
	ofn.lpstrFile = HeapAlloc(hHeap, 0, 512);
	ofn.lpstrFile[0] = 0;
	ofn.nMaxFile = 512;
	ofn.Flags |= OFN_CREATEPROMPT|OFN_OVERWRITEPROMPT;
	if (GetSaveFileName(&ofn) == FALSE)
	{
		HeapFree(hHeap, 0, ofn.lpstrFile);
		return FALSE;
	}
	lstrcpy(szBufferFilename, ofn.lpstrFile);
	HeapFree(hHeap, 0, ofn.lpstrFile);
	return TRUE;
}

BOOL GetLoadLifFilename(VOID)
{
	OPENFILENAME ofn;

	InitializeOFN(&ofn);
	ofn.lpstrFilter = _T("All Files (*.*)\0*.*\0") _T("\0\0");
	ofn.nFilterIndex = 1;
	ofn.lpstrFile = HeapAlloc(hHeap, 0, 512);
	ofn.lpstrFile[0] = 0;
	ofn.nMaxFile = 512;
	ofn.Flags |= OFN_FILEMUSTEXIST|OFN_PATHMUSTEXIST;
	if (GetOpenFileName(&ofn) == FALSE)
	{
		HeapFree(hHeap, 0, ofn.lpstrFile);
		return FALSE;
	}
	lstrcpy(szBufferFilename, ofn.lpstrFile);
	HeapFree(hHeap, 0, ofn.lpstrFile); 
	return TRUE;
}

BOOL GetSaveLifFilename(VOID)
{
	OPENFILENAME ofn;

	InitializeOFN(&ofn);
	ofn.lpstrFilter = _T("All Files (*.*)\0*.*\0") _T("\0\0");
	ofn.nFilterIndex = 1;
	ofn.lpstrFile = HeapAlloc(hHeap, 0, 512);
	ofn.lpstrFile[0] = 0;
	ofn.nMaxFile = 512;
	ofn.Flags |= OFN_CREATEPROMPT|OFN_OVERWRITEPROMPT;
	if (GetSaveFileName(&ofn) == FALSE)
	{
		HeapFree(hHeap, 0, ofn.lpstrFile);
		return FALSE;
	}
	lstrcpy(szBufferFilename, ofn.lpstrFile);
	HeapFree(hHeap, 0, ofn.lpstrFile); 
	return TRUE;
}

BOOL GetSaveDiskFilename(LPCTSTR lpstrName)
{
	OPENFILENAME ofn;

	InitializeOFN(&ofn);
	ofn.lpstrFilter = _T("All Files (*.*)\0*.*\0") _T("\0\0");
	ofn.nFilterIndex = 1;
	ofn.lpstrFile = HeapAlloc(hHeap, 0, 512);
	lstrcpy(ofn.lpstrFile, lpstrName);
	ofn.nMaxFile = 512;
	ofn.Flags |= OFN_CREATEPROMPT|OFN_OVERWRITEPROMPT;
	if (GetSaveFileName(&ofn) == FALSE)
	{
		HeapFree(hHeap, 0, ofn.lpstrFile);
		return FALSE;
	}
	lstrcpy(szBufferFilename, ofn.lpstrFile);
	HeapFree(hHeap, 0, ofn.lpstrFile); 
	return TRUE;
}


//################
//#
//#    Load and Save Lif files
//#
//################

static WORD Read2(LPBYTE adr)
{
	return (adr[0]<<8)|adr[1];
}

static WORD LifEndSpace(LPBYTE lif)
{
	WORD high;
	LPBYTE adr;

	DWORD dir_entry;

	high = Read2(lif + 0x0A) + Read2(lif + 0x12);
	adr = lif + (Read2(lif + 0x0A) << 8);
	dir_entry = Read2(lif + 0x12) * 8;
	do
	{
		if (Read2(adr + 0x0A) != 0xFFFF)
		{
			if ((Read2(adr + 0x0E) + Read2(adr + 0x12)) >= high)
				high = Read2(adr + 0x0E) + Read2(adr + 0x12);
		}
		adr += 0x20;
	} while (--dir_entry);
	return (WORD)(high);
}

static LPBYTE LifFreeEntry(LPBYTE lif)
{
	LPBYTE adr;
	WORD dir_entry;

	adr = lif + (Read2(lif + 0x0A) << 8);
	dir_entry = Read2(lif + 0x12) * 8;
	do
	{
		if (Read2(adr + 0x0A) == 0xFFFF)
			return adr;
		adr += 0x20;
	} while (--dir_entry);
	return 0;
}

static LPBYTE LifEntry(LPBYTE lif, WORD n)
{
	if (n < (8 * Read2(lif + 0x12)))
		return lif + (Read2(lif + 0x0A) << 8) + 0x20*n;
	return 0;
}

BOOL LoadLif(LPCTSTR szFilename)
{
	HANDLE hFile;
	DWORD  dwFileSizeLow;
	DWORD  dwFileSizeHigh;
	LPBYTE lpBuf;

	LPBYTE lif_src, lif_dest;
	WORD   freesect;

	if (!IsEdisk())
	{
		AbortMessage(_T("The computer does not have an\nedisk configured."));
		return FALSE;
	}

	hFile = CreateFile(szFilename, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (hFile == INVALID_HANDLE_VALUE) return FALSE;
	dwFileSizeLow = GetFileSize(hFile, &dwFileSizeHigh);
	if (dwFileSizeHigh != 0)
	{ // file is too large.
		CloseHandle(hFile);
		return FALSE;
	}
	lpBuf = (LPBYTE)HeapAlloc(hHeap, 0, dwFileSizeLow);
	if (lpBuf == NULL)
	{
		CloseHandle(hFile);
		return FALSE;
	}
	ReadFile(hFile, lpBuf, dwFileSizeLow, &dwFileSizeHigh, NULL);
	CloseHandle(hFile);

	lif_src = LifEntry(lpBuf, 0);
	lif_dest = LifFreeEntry(edStart);

	freesect = LifEndSpace(edStart);

	if ((edNsect - freesect) < Read2(lif_src + 0x12))
	{
		AbortMessage(_T("The computer does not have enough\nfree edisk to load this binary file."));
		HeapFree(hHeap, 0, lpBuf);
		return FALSE;
	}

	memcpy(edStart + (freesect << 8), lpBuf + (Read2(lif_src + 0x0E) << 8), Read2(lif_src + 0x12) << 8);

	lif_dest[0] = 'N'; lif_dest[1] = 'e';
	lif_dest[2] = 'w'; lif_dest[3] = ' ';
	lif_dest[4] = 'f'; lif_dest[5] = 'i';
	lif_dest[6] = 'l'; lif_dest[7] = 'e';
	lif_dest[8] = ' '; lif_dest[9] = ' ';

	memcpy(lif_dest + 0x0A, lif_src + 0x0A, 2);
	lif_dest[0x0E] = freesect >> 8;
	lif_dest[0x0F] = freesect & 0xFF;
	memcpy(lif_dest + 0x12, lif_src + 0x12, 2);
	memcpy(lif_dest + 0x1A, lif_src + 0x1A, 6);

	HeapFree(hHeap, 0, lpBuf);
	return TRUE;
}

BOOL SaveLif(LPCTSTR szFilename)			
{
	return TRUE;
}

BOOL LoadDisk(LPCTSTR szFilename)			// Load whole edisk as file
{
	HANDLE	hFile;
	DWORD	dwFileSizeLow;
	DWORD	dwFileSizeHigh;
	
	if (! IsEdisk())
	{
		AbortMessage(_T("The computer does not have an\nedisk configured."));
		return FALSE;
	}
	
	hFile = CreateFile(szFilename, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (hFile == INVALID_HANDLE_VALUE) return FALSE;
	dwFileSizeLow = GetFileSize(hFile, &dwFileSizeHigh);
	if (dwFileSizeHigh != 0)
	{ // file is too large.
		CloseHandle(hFile);
		return FALSE;
	}
	if ((dwFileSizeLow & 0x0000FF) != 0x00)
	{
		CloseHandle(hFile);
		AbortMessage(_T("The image is not a correct size"));
		return FALSE;
	}
	if (dwFileSizeLow > (DWORD)(edNsect << 8))
	{
		CloseHandle(hFile);
		AbortMessage(_T("The edisk configured on the computer is\nsmaller than the disk image."));
		return FALSE;
	}

	ReadFile(hFile, edStart, edNsect << 8, &dwFileSizeHigh, NULL);

	CloseHandle(hFile);

	edStart[0] = (edNsect) & 0x00FF;		// correct number of sectors after load
	edStart[1] = (edNsect >> 8) & 0x00FF;

	return TRUE;
}

BOOL SaveDisk(LPCTSTR szFilename)			// Save whole edisk as file
{
	HANDLE	hFile;
	DWORD	dwBytesWritten;
	
	if (! IsEdisk())
	{
		AbortMessage(_T("The computer does not have an\nedisk configured."));
		return FALSE;
	}

	hFile = CreateFile(szFilename, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
	{
		AbortMessage(_T("Cannot open file."));
		return FALSE;
	}

	WriteFile(hFile, edStart, edNsect << 8, &dwBytesWritten, NULL);
	CloseHandle(hFile);

	return TRUE;
}

//################
//#
//#    Load Bitmap
//#
//################

static __inline UINT DibNumColors (LPBITMAPINFOHEADER lpbi)
{
	if (lpbi->biClrUsed != 0) return (UINT)lpbi->biClrUsed;

	// a 24 bitcount DIB has no color table
	return (lpbi->biBitCount <= 8) ? (1 << lpbi->biBitCount) : 0;
}

static HPALETTE CreateBIPalette(LPBITMAPINFOHEADER lpbi)
{
	LOGPALETTE* pPal;
	HPALETTE    hpal = NULL;
	UINT        nNumColors;
	BYTE        red;
	BYTE        green;
	BYTE        blue;
	UINT        i;
	RGBQUAD*    pRgb;

	if (!lpbi)
		return NULL;

	if (lpbi->biSize != sizeof(BITMAPINFOHEADER))
		return NULL;

	// Get a pointer to the color table and the number of colors in it
	pRgb = (RGBQUAD FAR *)((LPBYTE)lpbi + (WORD)lpbi->biSize);
	nNumColors = DibNumColors(lpbi);

	if (nNumColors)
	{
		// Allocate for the logical palette structure
		pPal = (LOGPALETTE*)HeapAlloc(hHeap, 0, sizeof(LOGPALETTE) + nNumColors * sizeof(PALETTEENTRY));
		if (!pPal)
			return NULL;

		pPal->palNumEntries = nNumColors;
		pPal->palVersion    = 0x300;

		// Fill in the palette entries from the DIB color table and
		// create a logical color palette.
		for (i = 0; i < nNumColors; i++)
		{
			pPal->palPalEntry[i].peRed   = pRgb[i].rgbRed;
			pPal->palPalEntry[i].peGreen = pRgb[i].rgbGreen;
			pPal->palPalEntry[i].peBlue  = pRgb[i].rgbBlue;
			pPal->palPalEntry[i].peFlags = (BYTE)0;
		}
		hpal = CreatePalette(pPal);
		HeapFree(hHeap, 0, (HANDLE)pPal);
	}
	else
	{
		if (lpbi->biBitCount == 24)
		{
			// A 24 bitcount DIB has no color table entries so, set the
			// number of to the maximum value (256).
			nNumColors = 256;
			pPal = (LOGPALETTE*)HeapAlloc(hHeap, 0, sizeof(LOGPALETTE) + nNumColors * sizeof(PALETTEENTRY));
			if (!pPal)
				return NULL;

			pPal->palNumEntries = nNumColors;
			pPal->palVersion    = 0x300;

			red = green = blue = 0;

			// Generate 256 (= 8*8*4) RGB combinations to fill the palette
			// entries.
			for (i = 0; i < pPal->palNumEntries; i++)
			{
				pPal->palPalEntry[i].peRed   = red;
				pPal->palPalEntry[i].peGreen = green;
				pPal->palPalEntry[i].peBlue  = blue;
				pPal->palPalEntry[i].peFlags = (BYTE)0;

				if (!(red += 32))
					if (!(green += 32))
					blue += 64;
			}
			hpal = CreatePalette(pPal);
			HeapFree(hHeap, 0, (HANDLE)pPal);
		}
	}
	return hpal;
}

HBITMAP LoadBitmapFile(LPCTSTR szFilename)
{
	HANDLE  hFile;
	HANDLE  hMap;
	LPBYTE  pbyFile;
	HBITMAP hBitmap;
	LPBITMAPFILEHEADER pBmfh;
	LPBITMAPINFO pBmi;

//	if (hRomFile == NULL) return NULL;
	SetCurrentDirectory(szHP8xDirectory);
	hFile = CreateFile(szFilename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	SetCurrentDirectory(szCurrentDirectory);
	// opened with GENERIC_READ -> PAGE_READONLY
	hMap = CreateFileMapping(hFile, NULL, PAGE_READONLY, 0, 0, NULL);
	if (hMap == NULL)
	{
		CloseHandle(hFile);
		return NULL;
	}
	// opened with GENERIC_READ -> PAGE_READONLY -> FILE_MAP_READ
	pbyFile = MapViewOfFile(hMap, FILE_MAP_READ, 0, 0, 0);
	if (pbyFile == NULL)
	{
		CloseHandle(hMap);
		CloseHandle(hFile);
		return NULL;
	}

	hBitmap = NULL;
	pBmfh = (LPBITMAPFILEHEADER)pbyFile;
	if (pBmfh->bfType != 0x4D42) goto quit; // "BM"
	pBmi = (LPBITMAPINFO)(pbyFile+sizeof(BITMAPFILEHEADER));

	if (hPalette == NULL)
	{
		_ASSERT(hPalette == NULL);				// resource free
		hPalette = CreateBIPalette(&pBmi->bmiHeader);
		// save old palette
		hOldPalette = SelectPalette(hWindowDC, hPalette, FALSE);
		RealizePalette(hWindowDC);
	}
	if (FALSE&&(pBmi->bmiHeader.biBitCount <= 8))
	{
		UINT i;
		LPDWORD pdwTable = ((LPDWORD)&(pBmi->bmiColors[0]));
		if (pBmi->bmiHeader.biClrUsed)
			for (i=0; i<(UINT)pBmi->bmiHeader.biClrUsed; i++) pdwTable[i]=i;
		else
			for (i=0; i<256; i++) pdwTable[i]=i;
		hBitmap = CreateDIBitmap(
			hWindowDC,
			&pBmi->bmiHeader,
			CBM_INIT,
			pbyFile+pBmfh->bfOffBits,
			pBmi, DIB_PAL_COLORS);
    }
	else
	{
		hBitmap = CreateDIBitmap(
			hWindowDC,
			&pBmi->bmiHeader,
			CBM_INIT,
			pbyFile+pBmfh->bfOffBits,
			pBmi, DIB_RGB_COLORS);
	}
	_ASSERT(hBitmap != NULL);

quit:
	UnmapViewOfFile(pbyFile);
	CloseHandle(hMap);
	CloseHandle(hFile);
	return hBitmap;
}
