/*
 *   Hp-7470.c
 *
 *   Copyright 2004 Olivier De Smet
 */

//
//   HP7470A plotter as ADDRESS 5 (not done)
//

#include "StdAfx.h"
#include "HP8x.h"
#include "resource.h"

#include "math.h"

#if defined HP7470

#define HP7470ADR 0x5						// HP-IB address

#define DEBUG_HP7470

#define DEBUG_GL

#if defined DEBUG_HP7470
static WORD HP7470_ADR[] = {				// states of automata
	0, 1000, 1100, 1400, 
	10000
};

static TCHAR *HP7470_TXT[] ={				// labels of states of automata 
	_T("Idle"), _T("MSA received"), _T("MLA received"), _T("MTA received"),
	_T("Receive data")
};

#define NLAB (sizeof(HP7470_ADR)/sizeof(HP7470_ADR[0]))

static TCHAR *HP7470_LAB(WORD d)			// find a label for the state d
{
	WORD i;

	for(i = 0; i < NLAB; i++)
		if (HP7470_ADR[i] == d)
			return HP7470_TXT[i];
	return NULL;
}
#endif

// hp7470 variables

static WORD st7470 = 0;								// state of hp7470 controller
static BOOL hp7470_talk = FALSE;					// MTA received ?
static BOOL hp7470_listen = FALSE;					// MLA received ?
static TCHAR hp7470_file[MAX_PATH];					// file for printing
static WORD hp7470_fn = 0;							// number for file name
static HANDLE  hp7470_hfile = NULL;					// handle of file
static BYTE hc_7470[0x200];							// stack of commands
static WORD hc_7470_hi = 0;							// hi mark
static WORD hc_7470_lo = 0;							// lo mark
static BYTE	hd_7470[0x200];							// stack of data in
static WORD hd_7470_hi = 0;							// hi mark
static WORD hd_7470_lo = 0;							// lo mark

BOOL hp7470_ppol_e = TRUE;							// parallel poll enabled

// hp7470 emulator

#define HP7470MAXX 10900							// A4 paper
#define HP7470MAXY 7650

typedef struct {
	INT		X;										// current position
	INT		Y;
	INT		P1x;									// P1
	INT		P1y;
	INT		P2x;									// P2
	INT		P2y;
	INT		Xlo;									// drawing window
	INT		Xhi;
	INT		Ylo;
	INT		Yhi;
	INT		Xmin;									// scale parameters
	INT		Xmax;
	INT		Ymin;
	INT		Ymax;
	INT		Cangle;									// chord angle (AA, AR, ...)
	DOUBLE	Run;									// for label direction
	DOUBLE	Rise;
	DOUBLE	Tpx;									// for ticks length
	DOUBLE	Tpy;
	DOUBLE	Tnx;
	DOUBLE	Tny;
	DOUBLE	Cwidth;									// character width
	DOUBLE	Cheight;
	DOUBLE	Lplength;								// line pattern length
	DOUBLE	Cslant;									// character slant in degrees
	BYTE	Ltype;									// line type
	BYTE	Pen;									// number of pen
	BOOL	PenUp;									// state of pen
	BYTE	Emask;									// Error E mask
	BYTE	Smask;									//       P mask
	BYTE	Pmask;									//       S mask
	BYTE	Lterm;									// Label terminator
	BYTE	Pmode;									// plotting mode absolute, relative
	BYTE	Smode;									// symbol mode
	BYTE	Acs;									// alternate character set
	BYTE	Scs;									// standard character set
	BYTE	Cs;										// current character set selected
	BYTE	Scale;									// scale on off
	BYTE	Dclear;									// digitize clear
	BYTE	Error;									// error
	BYTE	Status;									// status bits
													// 0: pen down;
													// 1: P1 or P2 changed
													// 2: digitized point available
													// 3: plotter initialized
													// 4: ready for data
													// 5: error
} TRACER;

static TRACER trc;
static BYTE trc_buffer[1024];
static WORD trc_ptr = 0;

static BYTE trc_buffer_out[1024];
static WORD trc_ptr_out = 0;

static DWORD dwSheetColor[2] =						// color table for paper
{
	0x00FFFFFF, 0x000030C0
};

static struct
{
	BITMAPINFOHEADER Paper_bmih;
	RGBQUAD bmiColors[2]; 
} bmiSheet =
{
	{0x28,0/*x*/,0/*y*/,1,8,BI_RGB,0,0,0,2,0}
};
static  LPBYTE	pbySheet;				// address of bitmap for the paper

static	HBITMAP hSheetBitmap;			// paper bitmap

HWND	hWndTrc = NULL;					// handle of tracer window
HDC		hWindowDCTrc = NULL;			// handle of tracer window context
HDC		hSheetDC = NULL;				// handle of bitmap for sheet of paper context

UINT	nTrcX = 0;						// offset of printer X
UINT	nTrcY = 0;						// offset of printer Y
UINT	nTrcW = HP7470MAXX/10;					// width of visible paper
UINT	nTrcH = HP7470MAXY/10;					// height of visible paper

VOID hp_gl(VOID);

//################
//#
//#    Low level subroutines
//#
//################

BOOL hp7470_pop_c(BYTE *c)
{
	if (hc_7470_hi == hc_7470_lo)
		return FALSE;
	*c = hc_7470[hc_7470_lo++];
	hc_7470_lo &= 0x1FF;
	return TRUE;
}

BOOL hp7470_pop_d(BYTE *c)
{
	if (hd_7470_hi == hd_7470_lo)
		return FALSE;
	*c = hd_7470[hd_7470_lo++];
	hd_7470_lo &= 0x1FF;
	return TRUE;
}

BOOL hp7470_push_c(BYTE addr, BYTE c)	// push on stack hc_7470
{
	if (addr != HP7470ADR)
		return FALSE;
	hc_7470[hc_7470_hi++] = c;
	hc_7470_hi &= 0x1FF;
	_ASSERT(hc_7470_hi != hc_7470_lo);
	return TRUE;
}

BOOL hp7470_push_d(BYTE addr, BYTE d)	// push on stack hd_7470
{	
	if (addr != HP7470ADR || !hp7470_listen)
		return FALSE;
	hd_7470[hd_7470_hi++] = d;
	hd_7470_hi &= 0x1FF;
	_ASSERT(hd_7470_hi != hd_7470_lo);
	return TRUE;
}

//################
//#
//#    Public functions
//#
//################

VOID DoHP7470(VOID)
{
#if defined DEBUG_HP7470
WORD hp7470st;
#endif

BYTE c;
int i;
int l;

	#if defined DEBUG_HP7470
	hp7470st = st7470;
	#endif

	if (hc_7470_hi != hc_7470_lo)
	{
		hp7470_pop_c(&c);
		c &= 0x7F;						// remove parity
		switch(c)
		{
			case (0x60 + HP7470ADR):	// my secondary address
				_ASSERT(0);
				break;
			case (0x20 + HP7470ADR):	// my listen address
				hp7470_listen = TRUE;
//				hp7470_untalk = FALSE;
				st7470 = 10000;
				break;
			case (0x40 + HP7470ADR):	// my talk address
				hp7470_talk = TRUE;
//				hp7470_untalk = FALSE;
				st7470 = 1400;
				break;
			case 0x3F:					// unlisten
				hp7470_listen = FALSE;
//				hp7470_untalk = FALSE;
				st7470 = 0;
				break;
			case 0x5F:					// untalk
				hp7470_talk = FALSE;
//				hp7470_untalk = TRUE;
				st7470 = 0;
				break;
			default:					// unknown
//				hp7470_untalk = FALSE;
				break;
		}
	}

	switch (st7470)
	{
	// IDLE STATE
		case 0:
			break;
	// receive my talk address
		case 1400:
//			_ASSERT(0);
			if (hp7470_talk)
			{
				i = 0;
				while (trc_ptr_out > 0)
				{
					h_push(trc_buffer_out[i], 0x00);
					i--;
					trc_ptr_out--;
				}
				h_push(0x0d, 0x00);
				h_push(0x0a, 0x80);
			}
			st7470++;
			break;
		case 1401:
			break;
	// MLA: receive data
		case 10000:
			if (hp7470_listen)
			{
				if (hp7470_pop_d(&c))
				{
					if (hp7470_hfile != NULL)
					{
						WriteFile(hp7470_hfile, &c, 1, &l, NULL);
					}
					trc_buffer[trc_ptr++] = c;
					_ASSERT (trc_ptr < 1024);
					if (trc_ptr > 2)
					{
						if ((trc_buffer[0] == 'L') && (trc_buffer[1] == 'B') && (trc_buffer[trc_ptr-1] == trc.Lterm))
						{
							hp_gl();
							trc_ptr = 0;
						}
						else if (c == ';')
						{
							hp_gl();
							trc_ptr = 0;
						}
					}
					else if (c == ';')
					{
						hp_gl();
						trc_ptr = 0;
					}
				}
			}
			st7470++;
			break;
		case 10001:
			break;
		default:
			_ASSERT(0);
			break;
	}
	#if defined DEBUG_HP7470
	{
		TCHAR buffer[256];
		TCHAR *TXT = HP7470_LAB((WORD)(st7470));
		int k;
		unsigned long l;

		if (hp7470st != st7470)
		{
			if (TXT == NULL)
				k = wsprintf(buffer,_T("%04X: HP7470 : %u->%u\n"), Chipset.R[4]|(Chipset.R[5]<<8), hp7470st, st7470);
			else
				k = wsprintf(buffer,_T("%04X: HP7470 : %u->%u ( %s )\n"), Chipset.R[4]|(Chipset.R[5]<<8), hp7470st, st7470, TXT);
#if defined DEBUG_ONDISK
			WriteFile(hDebug, buffer, (unsigned long) (k), &l, NULL); 
#endif
			OutputDebugString(buffer);
		}
	}
	#endif
}

BOOL hp7470_save(VOID)
{
	if (hp7470_hfile != NULL)
	{
		CloseHandle(hp7470_hfile);
		hp7470_fn++;
	}
	return TRUE;
}

BOOL hp7470_eject(VOID)
{
	hp7470_save();
	return TRUE;
}

BOOL hp7470_widle(VOID)
{
	DWORD dwRefTime;

	SuspendDebugger();									// suspend debugger

	dwRefTime = timeGetTime();
	Chipset.idle = FALSE;
	Sleep(10);
	// wait for the Iddle state with 5 sec timeout
	while (((timeGetTime() - dwRefTime) < 5000L) && !(st7470 == 0))
		Sleep(10);

	if (st7470 == 0)									// not timeout, hp7470 is idle
		return FALSE;									// idle state
	else
		ResumeDebugger();								// timeout, resume to debugger

	return TRUE;										// hp7470 was busy
}

VOID hp7470_init(VOID)
{
	TCHAR szFilename[MAX_PATH];
	int k;

	st7470 = 0;								// state of hp9121 controller
	hp7470_talk = FALSE;					// MTA received ?
	hp7470_listen = FALSE;					// MLA received ?
	hp7470_ppol_e = TRUE;					// parallel poll enabled
	hc_7470_hi = 0;							// hi mark
    hc_7470_lo = 0;							// lo mark
	hd_7470_hi = 0;							// hi mark
	hd_7470_lo = 0;							// lo mark

	if (hp7470_hfile == NULL)
	{
		k = wsprintf(szFilename, _T("plotter-%03d.txt"), hp7470_fn);
		szFilename[k] = 0x00;
		hp7470_hfile = CreateFile(szFilename,
					   GENERIC_WRITE,
					   FILE_SHARE_READ,
					   NULL,
					   CREATE_ALWAYS,
					   FILE_FLAG_SEQUENTIAL_SCAN,
					   NULL);
		if (hp7470_hfile == INVALID_HANDLE_VALUE)
		{
			hp7470_hfile = NULL;
			return;
		}
	}
	if (hWndTrc == NULL)
	{
		Trc_window();
	}
}

VOID hp7470_stop(VOID)
{
	hp7470_save();

	if (hWndTrc != NULL)
	{
		DestroyWindow(hWndTrc);
		while (hWndTrc != NULL)
			Sleep(20);
		Trc_closew();
	}

}

//################
//#
//#    Message Handlers
//#
//################

static LRESULT Trc_NewPaper(VOID)
{
	BitBlt(hSheetDC, 0, 0, nTrcW, nTrcH, hSheetDC, 0, 0, BLACKNESS);
	EnterCriticalSection(&csGDILock);	// solving NT GDI problems
	{
		// redraw main display area
		BitBlt(hWindowDCTrc, nTrcX, nTrcY,
			   nTrcW, nTrcH,
			   hSheetDC, 0, 0, SRCCOPY);
		GdiFlush();
	}
	LeaveCriticalSection(&csGDILock); 
	return 0;
}

static LRESULT OnCreateTrc(HWND hWindow)
{
//	InitializeCriticalSection(&csGDILock);
//	InitializeCriticalSection(&csKeyLock);

	hWndTrc = hWindow;
	hWindowDCTrc = GetDC(hWndTrc);
	DragAcceptFiles(hWndTrc,FALSE);				// no support for dropped files
	Trc_NewPaper();

	return 0;
}

static LRESULT OnDestroyTrc(HWND hWindow)
{
	DragAcceptFiles(hWndTrc,FALSE);			// no WM_DROPFILES message any more
//	SwitchToState(SM_RETURN);				// exit emulation thread
//	SetWindowTitle(NULL);					// free memory of title
	ReleaseDC(hWndTrc, hWindowDCTrc);
	hWindowDCTrc = NULL;					// hWindowDC isn't valid any more
	hWndTrc = NULL;

//	DeleteCriticalSection(&csGDILock);
//	DeleteCriticalSection(&csKeyLock);

//	PostQuitMessage(0);					// exit message loop
	return 0;
	UNREFERENCED_PARAMETER(hWindow);
}

static LRESULT OnPaintTrc(HWND hWindow)
{
	PAINTSTRUCT Paint;
	HDC hPaintDC;

//	int pos;

	hPaintDC = BeginPaint(hWindow, &Paint);
	if (hWindowDCTrc != NULL)
	{
		EnterCriticalSection(&csGDILock);	// solving NT GDI problems
		{
			// redraw main display area
			BitBlt(hWindowDCTrc, nTrcX, nTrcY, nTrcW, nTrcH,
				   hSheetDC, 0, 0, SRCCOPY);
			GdiFlush();
		}
		LeaveCriticalSection(&csGDILock); 
	}
	EndPaint(hWindow, &Paint);
	return 0;
}

LRESULT CALLBACK TrcWndProc(HWND hWindow, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_CREATE:          return OnCreateTrc(hWindow);
	case WM_DESTROY:         return OnDestroyTrc(hWindow);
	case WM_PAINT:           return OnPaintTrc(hWindow);
	case WM_ACTIVATE:
		if (LOWORD(wParam)==WA_INACTIVE) break;
	case WM_QUERYNEWPALETTE:
		if (hPalette)
		{
			SelectPalette(hWindowDC, hPalette, FALSE);
			if (RealizePalette(hWindowDCTrc))
			{
				InvalidateRect(hWindow, NULL, TRUE);
				return TRUE;
			}
		}
		return FALSE;
	case WM_PALETTECHANGED:
		if ((HWND)wParam == hWindow) break;
		if (hPalette)
		{
			SelectPalette(hWindowDC, hPalette, FALSE);
			if (RealizePalette(hWindowDCTrc))
			{
				// UpdateColors(hWindowDC);
				InvalidateRect (hWnd, (LPRECT) (NULL), 1);
			}
		}
		return FALSE; 
/*	case WM_VSCROLL:
		switch (LOWORD(wParam))
		{
		case SB_LINEUP:		return OnScroll(GetScroll()-18);
		case SB_LINEDOWN:	return OnScroll(GetScroll()+18);
		case SB_PAGEUP:		return OnScroll(GetScroll()-((INT)nPrt85H));
		case SB_PAGEDOWN:	return OnScroll(GetScroll()+((INT)nPrt85H));
		case SB_THUMBTRACK:
			GetScroll();
			return OnScroll(nPrt85H + (nHeadY - nPrt85H) * HIWORD(wParam) / (10000 - sPrt85.nPage));
		}
		break;*/
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case ID_PAPER_NEW:		return Trc_NewPaper();
		case ID_PAPER_SETTINGS:	return 0;
		case ID_PAPER_SAVEAS:	return 0;
		}
		break;
	case WM_SYSCOMMAND:
		switch (LOWORD(wParam))
		{
		case SC_CLOSE: return /* OnFileExit() */ 0;
		}
		break;
	}
	return DefWindowProc(hWindow, uMsg, wParam, lParam);
}

VOID Trc_closew(VOID)
{
	if (hSheetDC != NULL)
	{
		SelectObject(hSheetDC, hSheetBitmap);
		DeleteObject(hSheetBitmap);
		DeleteDC(hSheetDC);
		hSheetDC = NULL;
		hSheetBitmap = NULL;
	}

	if ( !UnregisterClass(_T("CHP8xT"), hApp))
	{
		AbortMessage(
			_T("CHP8xT class unregistration failed.\n")
			_T("This application will now terminate."));
		return;
	}
}

VOID Trc_window(VOID)
{
	RECT rectWindow;
	WNDCLASS wc;

	wc.style = CS_BYTEALIGNCLIENT;
	wc.lpfnWndProc = (WNDPROC)TrcWndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hApp;
	wc.hIcon = LoadIcon(hApp, MAKEINTRESOURCE(IDI_HP8X));
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = NULL;
	wc.lpszMenuName = MAKEINTRESOURCE(IDR_MENU_TRC);
	wc.lpszClassName = _T("CHP8xT");

	if (!RegisterClass(&wc))
	{
		AbortMessage(
			_T("CHP8xT class registration failed.\n")
			_T("This application will now terminate."));
		return;
	}

	// Create window
	rectWindow.left   = 0;
	rectWindow.top    = 0;
	rectWindow.right  = nTrcW;
	rectWindow.bottom = nTrcH;
	AdjustWindowRect(&rectWindow, WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX|WS_OVERLAPPED, TRUE);

	hWndTrc = CreateWindow(_T("CHP8xT"), _T("HP8x Tracer"),
		WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX|WS_OVERLAPPED,
		CW_USEDEFAULT, CW_USEDEFAULT,
		rectWindow.right  - rectWindow.left,
		rectWindow.bottom - rectWindow.top,
		NULL, NULL, hApp, NULL
		);

	if (hWndTrc == NULL)
	{
		AbortMessage(_T("Window creation failed."));
		return;
	}

	// create printer bitmaps
	bmiSheet.Paper_bmih.biWidth = nTrcW;
	bmiSheet.Paper_bmih.biHeight = -(INT)(nTrcH);
	bmiSheet.bmiColors[1] = *(RGBQUAD*)&dwSheetColor[1];    // pixel on  color
	bmiSheet.bmiColors[0] = *(RGBQUAD*)&dwSheetColor[0];	// pixel off color

	_ASSERT(hSheetDC == NULL);
	hSheetDC = CreateCompatibleDC(hWindowDCTrc);
	_ASSERT(hSheetDC != NULL);
	hSheetBitmap = CreateDIBSection(hSheetDC, (BITMAPINFO*)&bmiSheet,DIB_RGB_COLORS, (LPVOID*)&pbySheet, NULL, 0);
	_ASSERT(hSheetBitmap != NULL);
	hSheetBitmap = SelectObject(hSheetDC, hSheetBitmap);

	SetDIBColorTable(hSheetDC,0,ARRAYSIZEOF(bmiSheet.bmiColors),bmiSheet.bmiColors); 

	ShowWindow(hWndTrc, SW_SHOWNA);
}

//################
//#
//#    HP-GL emulator
//#
//################

VOID gl_push_outc(BYTE car)
{
	trc_buffer_out[trc_ptr_out++] = car;
	_ASSERT(trc_ptr_out < 1024);
}

VOID gl_push_outs(char *str, int k)
{
INT i = 0;

	while (k)
	{
		k--;
		gl_push_outc(str[i++]);
	}
}

BOOL gl_next_val(WORD *ptr)
{
WORD pos = *ptr;
BOOL got = FALSE;

	do
	{
		if ((trc_buffer[pos] >= '0') && (trc_buffer[pos] <= '9'))
		{
			got = TRUE;
			break;
		}
		if (trc_buffer[pos] == '-')
		{
			got = TRUE;
			break;
		}
		if (trc_buffer[pos] == '+')
		{
			got = TRUE;
			break;
		}
		if (trc_buffer[pos] == '.')
		{
			got = TRUE;
			break;
		}
		if (trc_buffer[pos] == ';')
			break;
		pos++;
		if (pos == trc_ptr)
			break;
	} while (1);
	if (pos == trc_ptr)
		return FALSE;
	if (got)
		*ptr = pos;
	return got;
}

BOOL gl_getint(WORD *ptr, INT *val)	// get an integer param
{
WORD pos;
BOOL got = FALSE;
BOOL neg = FALSE;

	if (!gl_next_val(ptr))
		return FALSE;

	pos = *ptr;
	*val = 0;
	if (trc_buffer[pos] == '-')
	{
		neg = TRUE;
		pos++;
	}
	else if (trc_buffer[pos] == '+')
	{
		neg = FALSE;
		pos++;
	}

	while ((trc_buffer[pos] >= '0') && (trc_buffer[pos] <= '9'))
	{
		*val = (*val * 10)  + (trc_buffer[pos++] - '0');
		got = TRUE;
	}
	if (neg)
		*val = - (*val);
	if (got)
		*ptr = pos;
	return got;
}

BOOL gl_getdeci(WORD *ptr, DOUBLE *val)	// get a numerical param
{
WORD pos;
DOUBLE dec = 10.0;
BOOL got = FALSE;
BOOL neg = FALSE;

	if (!gl_next_val(ptr))
		return FALSE;

	pos = *ptr;
	*val = 0.0;

	if (trc_buffer[pos] == '-')
	{
		neg = TRUE;
		pos++;
	}
	else if (trc_buffer[pos] == '+')
	{
		neg = FALSE;
		pos++;
	}

	while ((trc_buffer[pos] >= '0') && (trc_buffer[pos] <= '9'))
	{
		*val = (*val * 10.0)  + (DOUBLE)(trc_buffer[pos++] - '0');
		got = TRUE;
	}
	if (trc_buffer[pos] == '.')
	{
		pos++;
		while ((trc_buffer[pos] >= '0') && (trc_buffer[pos] <= '9'))
		{
			*val += (DOUBLE)(trc_buffer[pos++] - '0')/dec;
			dec *= 10.0;
			got = TRUE;
		}
	}
	if (neg)
		*val = - (*val);
	if (got)
		*ptr = pos;
	return got;
}

BOOL gl_getdeci16(WORD *ptr, INT *val)	// get a 16 bits signed numerical param
{
DOUBLE param;

	if (gl_getdeci(ptr, &param))
	{
		*val = (INT)(floor(param));
		if ((*val < -32768) || (*val > 32767))
			trc.Error = 3;
		return TRUE;
	}
	return FALSE;
}

DOUBLE gl_dist(DOUBLE x1, DOUBLE y1, DOUBLE x2, DOUBLE y2)
{
	return sqrt((x2-x1) * (x2-x1) + (y2-y1) * (y2-y1));
}

VOID gl_DF(WORD pos)				// DF									Set default values
{
DOUBLE param;

	if (gl_getdeci(&pos, &param))
	{
		trc.Error = 2;
		return;
	}
	
	trc.Pmode = 1;							// absolute plotting mode
	trc.Run = (trc.P2x - trc.P1x) / 100.0;	// DR 1, 0
	trc.Rise = 0.0;
	trc.Ltype = 1;							// line type solid (1)
	trc.Lplength = 0.04 * gl_dist(trc.P1x, trc.P1y, trc.P2x, trc.P2y);
	trc.Xlo = 0;							// input window
	trc.Xhi = HP7470MAXX;
	trc.Ylo = 0;
	trc.Yhi = HP7470MAXY;
	trc.Cwidth = 0.0075 * (trc.P2x - trc.P1x);
	trc.Cheight = 0.005 * (trc.P2y - trc.P1y);
	trc.Smode = 0;							// symbol mode off
	trc.Tnx = trc.Tpx = 0.005 * (trc.P2x - trc.P1x);
	trc.Tny = trc.Tpy = 0.005 * (trc.P2y - trc.P1y);
	trc.Scs = 0;							// standard character set
	trc.Acs = 0;							// alternate character set
	trc.Cs = 0;								// standard character set selected
	trc.Cslant = 0;							// slant
	trc.Emask = 223;						// e,d,s masks
	trc.Smask = 0;
	trc.Pmask = 0;
	trc.Dclear = 1;							// digitize clear on
	trc.Scale = 0;							// scale off
	trc.Lterm = 0x03;						// label terminator
	trc.Cangle = 5;							// chord angle 5 degrees

#if defined DEBUG_GL
	{
		TCHAR buffer[256];
		int k;

		k = wsprintf(buffer,_T("HP7470 : got DF\n"));
#if defined DEBUG_ONDISK
		WriteFile(hDebug, buffer, (unsigned long) (k), &l, NULL); 
#endif
		OutputDebugString(buffer);
	}
#endif
}

VOID gl_IN(WORD pos)				// IN									Initialize
{
DOUBLE param;

	if (gl_getdeci(&pos, &param))
	{
		trc.Error = 2;
		return;
	}

	trc.P1x = 250;							// P1
	trc.P1y = 279;
	trc.P2x = 10250;						// P2
	trc.P2y = 7479;
	trc.Error = 0;
	trc.Status = 0x08;						// initialize ok

	gl_DF(pos);

#if defined DEBUG_GL
	{
		TCHAR buffer[256];
		int k;

		k = wsprintf(buffer,_T("HP7470 : got IN\n"));
#if defined DEBUG_ONDISK
		WriteFile(hDebug, buffer, (unsigned long) (k), &l, NULL); 
#endif
		OutputDebugString(buffer);
	}
#endif

}

VOID gl_IM(WORD pos)				// IM e(, s(, p)))						Input e, s and p masks
{
INT param;

	if (gl_getdeci16(&pos, &param))
	{
		trc.Emask = (BYTE)(param);
		if (gl_getdeci16(&pos, &param))
		{
			trc.Smask = (BYTE)(param);
			if (gl_getdeci16(&pos, &param))
			{
				trc.Pmask = (BYTE)(param);
			}
			else
			{
				trc.Pmask = 0;
			}
		}
		else
		{
			trc.Smask = 0;
			trc.Pmask = 0;
		}
	}
	else
	{
		trc.Emask = 223;
		trc.Smask = 0;
		trc.Pmask = 0;
	}
#if defined DEBUG_GL
	{
		TCHAR buffer[256];
		int k;

		k = wsprintf(buffer,_T("HP7470 : got IM %d, %d, %d\n"), (int)(trc.Emask), (int)(trc.Smask), (int)(trc.Pmask));
#if defined DEBUG_ONDISK
		WriteFile(hDebug, buffer, (unsigned long) (k), &l, NULL); 
#endif
		OutputDebugString(buffer);
	}
#endif
}

VOID gl_IP(WORD pos)				// IP P1x, P1y(, P2x, P2y)				Input P1 and P2
{
INT x1, y1;

	if (gl_getdeci16(&pos, &x1))
	{
		if (trc.Error)
			return;
		if (gl_getdeci16(&pos, &y1))
		{
			if (trc.Error)
				return;
			trc.P1x = x1;
			trc.P1y = y1;

			if (gl_getdeci16(&pos, &x1))
			{
				if (trc.Error)
					return;
				if (gl_getdeci16(&pos, &y1))
				{
					if (trc.Error)
						return;
					trc.P2x = x1;
					trc.P2y = y1;
				}
				else
				{
					trc.Error = 2;
					return;
				}
			}
		}
		else
		{
			trc.Error = 2;
			return;
		}
	}
	else
	{
		trc.P1x = 250;							// P1
		trc.P1y = 279;
		trc.P2x = 10250;						// P2
		trc.P2y = 7479;
	}
#if defined DEBUG_GL
	{
		TCHAR buffer[256];
		int k;

		k = wsprintf(buffer,_T("HP7470 : got IP %d, %d, %d, %d\n"), trc.P1x, trc.P1y, trc.P2x, trc.P2y);
#if defined DEBUG_ONDISK
		WriteFile(hDebug, buffer, (unsigned long) (k), &l, NULL); 
#endif
		OutputDebugString(buffer);
	}
#endif
}

VOID gl_OE(WORD pos)					// OE									Output error
{
TCHAR buff[256];
int k;

	k = wsprintf(buff, _T("%d"), (INT)(trc.Error));
	gl_push_outs(buff, k);
	trc.Status &= 0xEF;							// clear bit 5


#if defined DEBUG_GL
	{
		TCHAR buffer[256];
		int k;

		k = wsprintf(buffer,_T("HP7470 : got OE, send %d\n"), trc.Error);
#if defined DEBUG_ONDISK
		WriteFile(hDebug, buffer, (unsigned long) (k), &l, NULL); 
#endif
		OutputDebugString(buffer);
	}
#endif
}


VOID hp_gl(VOID)
{
	switch(trc_buffer[0])
	{
	case 'A':
		switch (trc_buffer[1])
		{
		case 'A':					// AA X,Y, arc angle (,chord angle)		Arc absolute
			break;
		case 'R':					// AR X,Y, arc angle (,chord angle)		Arc relative
			break;
		}
		break;
	case 'C':
		switch (trc_buffer[1])
		{
		case 'A':					// CA n									Designate alternate set n
			break;
		case 'I':					// CI radius (,chord angle)				Circle
			break;
		case 'P':					// CP spaces, lines						Character plot
			break;
		case 'S':					// CS m									Designate standard set m
			break;
		}
		break;
	case 'D':
		switch (trc_buffer[1])
		{
		case 'C':					// DC									Digitize clear
			break;
		case 'F':					// DF									Set default values
			gl_DF(2);
			break;
		case 'I':					// DI run, rise							Absolute direction
			break;
		case 'P':					// DP									Digitize point
			break;
		case 'R':					// DR run, rise							Relative direction
			break;
		case 'T':					// DT c									Define label terminator
			break;
		}
		break;
	case 'I':
		switch (trc_buffer[1])
		{
		case 'M':					// IM e(, s(, p)))						Input e, s and p masks
			gl_IM(2);
			break;
		case 'N':					// IN									Initialize
			gl_IN(2);
			break;
		case 'P':					// IP P1x, P1y(, P2x, P2y)				Input P1 and P2
			gl_IP(2);
			break;
		case 'W':					// IW Xlo, Ylo, Xhi, Yhi				Input window
			break;
		}
		break;
	case 'L':
		switch (trc_buffer[1])
		{
		case 'B':					// LB c...c								Label ASCII string
			break;
		case 'T':					// LT t(, l)							Designate line type and length
			break;
		}
		break;
	case 'O':
		switch (trc_buffer[1])
		{
		case 'A':					// OA									Output actual position and pen status
			break;
		case 'C':					// OC									Output commandes position and pen status
			break;
		case 'D':					// OD									Output digitized point and pen status
			break;
		case 'E':					// OE									Output error
			gl_OE(2);
			break;
		case 'F':					// OF									Output factors
			break;
		case 'I':					// OI									Output identification
			break;
		case 'O':					// OO									Output options
			break;
		case 'P':					// OP									Output P1 and P2
			break;
		case 'S':					// OS									Output status
			break;
		case 'W':					// OW									Output window
			break;
		}
		break;
	case 'P':
		switch (trc_buffer[1])
		{
		case 'A':					// PA x, y(, x, y(,...))				Plot absolute
			break;
		case 'D':					// PD (x, y(,...))						Pen down
			break;
		case 'R':					// PR x, y(, x, y(,...))				Plot relative
			break;
		case 'U':					// PU (x, y(,...))						Pen up
			break;
		}
		break;
	case 'S':
		switch (trc_buffer[1])
		{
		case 'A':					// SA									Select alternate character set
			break;
		case 'C':					// SC Xmin, Xmax, Ymin, Ymax			Scale
			break;
		case 'I':					// SI width, height						Absolute character size
			break;
		case 'L':					// SL tan theta							Absolute character slant (from vertical)
			break;
		case 'M':					// SM c									Symbol mode c
			break;
		case 'P':					// SP n									Select pen n
			break;
		case 'R':					// SR width, height						Relative character size
			break;
		case 'S':					// SS									Select standard character set
			break;
		}
		break;
	case 'T':
		switch (trc_buffer[1])
		{
		case 'L':					// TL tp(, tn)							Tick length
			break;
		}
		break;
	case 'U':
		switch (trc_buffer[1])
		{
		case 'C':					// UC (pen,) x, y, pen(,...)			User defined chracter
			break;
		}
		break;
	case 'V':
		switch (trc_buffer[1])
		{
		case 'S':					// VS v									Select velocity v
			break;
		}
		break;
	case 'X':
		switch (trc_buffer[1])
		{
		case 'T':					// XT									X axis tick
			break;
		}
		break;
	case 'Y':
		switch (trc_buffer[1])
		{
		case 'T':					// YT									Y axis tick
			break;
		}
		break;
	}
}

#endif