/*
 *   Serial.c
 *
 *   Copyright 2004 Olivier De Smet
 */

//
//   serial HP82939A as ADDRESS 0
//   not finished, have to look at serial on Windows first
//

#include "StdAfx.h"
#include "HP8x.h"
#include "io.h"								// I/O definitions

//#define DEBUG_SERIAL

static WORD SERIAL_ADR[] = {
	0, 1000, 1100, 1400, 
	2000,
	10000, 10500, 11000, 11500,
	12000, 12500, 13000, 13500,
	14000, 14500, 15000, 15500,
	16000, 16500, 17000, 17500,
	18000, 18500, 19000, 
	20000, 20500, 21000, 21500,
	22000, 22500
};

static TCHAR *SERIAL_TXT[] ={
	_T("Idle"), _T("MSA received"), _T("MLA received"), _T("MTA received"),
	_T("Identify"),
	_T("Receive data"), _T("Cold load read"), _T("Seek"), _T("Request status"),
	_T("Verify"), _T("Initialize"), _T("Request logical address"), _T("End"),
	_T("buffered write ?"), _T("Buffered read ?"), _T("unbuffered read verify"), _T("request physical address"),
	_T("Send wear"), _T("Format"), _T("Download"), _T("HP-300 clear"),
	_T("HP-IB-CRC L"), _T("Write loopback record"), _T("Initiate selftest"), 
	_T("Send data"), _T("Send status or address"), _T("DSJ"), _T("HP-IB-CRC T"),
	_T("Read loopback record"), _T("Read selftest")
};

#define NLAB (sizeof(SERIAL_ADR)/sizeof(SERIAL_ADR[0]))

static TCHAR *SERIAL_LAB(WORD d)
{
	WORD i;

	for(i = 0; i < NLAB; i++)
		if (SERIAL_ADR[i] == d)
			return SERIAL_TXT[i];
	return NULL;
}

// serial bus

static HANDLE hSerial;
static DCB    dcbSerial;

static WORD s_out[0x200];					// stack of data out of controler (back to computer)
WORD s_out_hi = 0;				
WORD s_out_lo = 0;

BOOL s_push(BYTE b, BYTE st)				// push on data out and status st = 0x80 : EOI
{
	s_out[s_out_hi++] = (WORD)(b | (st << 8));
	s_out_hi &= 0x1FF;
	_ASSERT(s_out_hi != s_out_lo);
	return TRUE;
}

static BOOL s_pop(BYTE *b, BYTE *st)							// pop on data out
{
	if (s_out_hi == s_out_lo)
		return FALSE;
	*b = (BYTE)(s_out[s_out_lo] & 0x0FF);
	*st = (BYTE)(s_out[s_out_lo++] >> 8);
	s_out_lo &= 0x1FF;
	return TRUE;
}

// hp82939 functions

static WORD st82939 = 0;								// state of hp9121 controller
static BYTE hc_82939[0x200];							// stack of commands
static WORD hc_82939_hi = 0;							// hi mark
static WORD hc_82939_lo = 0;							// lo mark
static BYTE	hd_82939[0x200];							// stack of data in
static WORD hd_82939_hi = 0;							// hi mark
static WORD hd_82939_lo = 0;							// lo mark

BOOL hp82939_widle(VOID)
{
	DWORD dwRefTime;

	SuspendDebugger();									// suspend debugger

	dwRefTime = timeGetTime();
	Chipset.idle = FALSE;
	Sleep(10);
	// wait for the Iddle state with 5 sec timeout
	while (((timeGetTime() - dwRefTime) < 5000L) && !(st82939 == 1))
		Sleep(10);

	if (st82939 == 1)									// not timeout, hp82939 is idle
		return FALSE;									// idle state
	else
		ResumeDebugger();								// timeout, resume to debugger

	return TRUE;										// hp82939 was busy
}

VOID hp82939_init(VOID)
{
	st82939 = 0;								// state of hp82939 controller
	hc_82939_hi = 0;							// hi mark
    hc_82939_lo = 0;							// lo mark
	hd_82939_hi = 0;							// hi mark
	hd_82939_lo = 0;							// lo mark
}


BOOL hp82939_pop_c(BYTE *c)
{
	if (hc_82939_hi == hc_82939_lo)
		return FALSE;
	*c = hc_82939[hc_82939_lo++];
	hc_82939_lo &= 0x1FF;
	return TRUE;
}

BOOL hp82939_pop_d(BYTE *c)
{
	if (hd_82939_hi == hd_82939_lo)
		return FALSE;
	*c = hd_82939[hd_82939_lo++];
	hd_82939_lo &= 0x1FF;
	return TRUE;
}

BOOL hp82939_push_c(BYTE c)	// push on stack hc_82939
{
	hc_82939[hc_82939_hi++] = c;
	hc_82939_hi &= 0x1FF;
	_ASSERT(hc_82939_hi != hc_82939_lo);
	return TRUE;
}

BOOL hp82939_push_d(BYTE d)	// push on stack hd_82939
{
	hd_82939[hd_82939_hi++] = d;
	hd_82939_hi &= 0x1FF;
	_ASSERT(hd_82939_hi != hd_82939_lo);
	return TRUE;
}

VOID DoHP82939(VOID)
{
#if defined DEBUG_SERIAL
WORD serialst;
#endif

//BYTE c;
//WORD d;

	#if defined DEBUG_SERIAL
	serialst = st82939;
	#endif

	switch (st82939)
	{
	// IDLE STATE
		case 0:
			break;
		case 1:
			break;
	}
	#if defined DEBUG_SERIAL
	{
		TCHAR buffer[256];
		TCHAR *TXT = SERIAL_LAB((WORD)(st82939));
		int k;
		unsigned long l;

		if (serialst != st82939)
		{
			if (TXT == NULL)
				k = wsprintf(buffer,_T("%04X: SERIAL : %u->%u\n"), Chipset.R[4]|(Chipset.R[5]<<8), serialst, st82939);
			else
				k = wsprintf(buffer,_T("%04X: SERIAL : %u->%u ( %s )\n"), Chipset.R[4]|(Chipset.R[5]<<8), serialst, st82939, TXT);
#if defined DEBUG_ONDISK
			WriteFile(hDebug, buffer, (unsigned long) (k), &l, NULL); 
//			FlushFileBuffers(hDebug);
#endif
			OutputDebugString(buffer);
		}
	}
	#endif
}

// function on serial

BOOL serial_receive(BYTE *b, BYTE *st)
{
	return s_pop(b, st);
}

BOOL serial_send_d(BYTE d, BYTE st) 
{
	hp82939_push_d(d);
	return TRUE;
}

BOOL serial_send_c(BYTE c) 
{
	hp82939_push_c(c);
	return TRUE;
}

BOOL serial_reset(IOP *i)
{

	if (hSerial)
		CloseHandle(hSerial);

	hSerial = CreateFile(_T("\\\\ .\\COM1"), GENERIC_READ|GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);
	if (hSerial == INVALID_HANDLE_VALUE)
		AbortMessage(_T("COM1: open error !"));

	GetCommState(hSerial, &dcbSerial);

	i->ctreg[0] = 0x00;		// CR0
	i->ctreg[1] = 0x00;		// CR1
	i->ctreg[2] = 0x00;		// CR2
	i->ctreg[3] = 0x06;		// CR3
	i->ctreg[4] = 0x00;		// CR4
	i->ctreg[5] = 0x00;		// CR5
	i->ctreg[6] = 0x00;		// CR6
	i->ctreg[7] = 0x00;		// CR7
	i->ctreg[8] = 0x00;		// CR8
	i->ctreg[9] = 0x89;		// CR9
	i->ctreg[11] = 0x00;	// CR11
	i->ctreg[12] = 0x00;	// CR12
	i->ctreg[13] = 0x00;	// CR13
	i->ctreg[14] = 0x00;	// CR14
	i->ctreg[15] = 0x00;	// CR15
	i->ctreg[16] = 0x02;	// CR16
	i->ctreg[17] = 0x0D;	// CR17
	i->ctreg[18] = 0x0A;	// CR18
	i->ctreg[19] = 0x00;	// CR19
	i->ctreg[20] = 0x00;	// CR20
	i->ctreg[21] = 0x00;	// CR21
	i->ctreg[22] = 0x00;	// CR22
	i->ctreg[23] = 0x00;	// CR23

	i->streg[0] = 0x02;		// SR0
	i->streg[1] = 0x00;		// SR1
	i->streg[2] = 0x00;		// SR2
	i->streg[3] = 0x00;		// SR3
	i->streg[4] = 0x00;		// SR4
	i->streg[5] = 0x00;		// SR5
	i->streg[6] = 0x00;		// SR6
	i->streg[7] = 0x00;		// SR7
	i->streg[8] = 0x00;		// SR8
	i->streg[9] = 0x89;		// SR9
	i->streg[10] = 0x00;	// SR10
	i->streg[11] = 0x00;	// SR11

	hp82939_init();

	return TRUE;
}

VOID serial_stop(IOP *i)
{
}

static VOID ifc0(IOP *i)		// ABORT I/O
{
	_ASSERT(0);
}

static VOID ifc1(IOP *i)		// REN = 1
{
	_ASSERT(0);
}

static VOID ifc2(IOP *i)		// REN = 0
{
	_ASSERT(0);
}

static VOID ifc3(IOP *i)		// ATN = 0
{
	_ASSERT(0);
}

static VOID ifc4(IOP *i)		// Perform Parallel Pool
{
	_ASSERT(0);
}

static VOID ifc5(IOP *i)		// Send "MY TALK ADDRESS"
{
	_ASSERT(0);
}

static VOID ifc6(IOP *i)		// Send "MY LISTEN ADDRESS"
{
	_ASSERT(0);
}

//
// not like this, have to repeat cr23 ..
//
static VOID ifc7(IOP *i)		// Send EOL character sequence
{
BYTE j, k;

	j = i->ctreg[16] & 0x07;
	k = 17;
	while( j > 0)
	{
		serial_send_d(i->ctreg[k++], 0);
		j--;
	}
}

static VOID ifc8(IOP *i)		// BREAK I/O
{
	_ASSERT(0);
}

static VOID ifc9(IOP *i)		// RESUME I/O
{
	_ASSERT(0);
}

VOID *serial_ifc[10] = {
	ifc0, ifc1, ifc2,
	ifc3, ifc4, ifc5, 
	ifc6, ifc7, ifc8, 
	ifc9
};
