/*
 *   Sound.c
 *
 *   Copyright 2004 Olivier De Smet
 */

//
//   sound for 1.2 Khz or variable one (8 bits mono 22050 Hz sound)
//
//   use Wave API and use a fade out hack to stop the sound in a nice (?) way
//   to avoid 'clicks'
//

#include "StdAfx.h"
#include "HP8x.h"
#include <math.h>

//#define DEBUG_SND					// flag for debug

#define SAMPLES 2205L				// base samples (0.1 sec)
#define SAMPLEFREQ 22050L			// sample freq 22.05 kHz

union PTR {							// struct to use quad words
	DWORD dw;
	struct {
		WORD lw;
		WORD hw;
	};
};

typedef struct {					// struct for handling windows wave 
	HANDLE hData;
	BYTE *lpData;
	HGLOBAL hWaveHdr;
	LPWAVEHDR lpWaveHdr;
} WS;

static WS wsSound12;				// sound for 1.2 khz
static WS wsSoundxx;				// variable sound
static HWAVEOUT hWaveOut;			// handle to the wave out device
static BOOL	bWaveOk = FALSE;		// init ok ? (for local use)
static BYTE bySnref[SAMPLES];		// reference sinus signal
static DWORD dwCfreq;				// current wsSoundxx frequency
static union PTR VolumeInit;		// initial volume level
static union PTR Volume;			// current volume level (used for stopping sound) 

DWORD dwSoundTicks;					// for counting Capricorn cycles between High and Low state of speaker
BOOL bInSound12 = FALSE;			// is Sound12 playing ?
BOOL bInSoundxx = FALSE;			// is Soundxx playing ?
BOOL bSoundStop = FALSE;			// to stop the sound in capricorn.c

//################
//#
//#    Low level subroutines
//#
//################

//
// Create reference sinus wave
//
static VOID SetSnRef(VOID)
{
	double drHz = 2.0*3.1415927/(double)(SAMPLES);
	double drAn = 0.0;
	int i;
	BYTE *lpAd = bySnref;

	for(i=0; i < SAMPLES; i++)
	{
		*lpAd++ = (BYTE)(128.0 + 127.0 * sin(drAn));
		drAn += drHz;
	}
}

//
//	Create lpAD sample for a sound at dwFreq Hz with iSamples samples
//
static VOID SetWaveFreq(BYTE *lpAd, DWORD dwFreq, int iSamples)
{
	DWORD dwHz = dwFreq*65536L;
	int i;
	union PTR an;

#if defined DEBUG_SND
	TCHAR buffer[256];

	wsprintf(buffer,_T("New Freq : %10ld \n"), dwFreq);
	OutputDebugString(buffer);
#endif

	_ASSERT(dwFreq < SAMPLEFREQ);

	an.dw = 0;
	dwHz /= 10L;
	for(i=0; i < iSamples; i++)
	{
		*lpAd++ = bySnref[an.hw];
		an.dw += dwHz;
		if (an.hw >= SAMPLES)
			an.hw -= SAMPLES;
	}
}

//
// Get and display some device capabilities (only for debugging)
//
#if defined DEBUG_SND
static VOID IdentifyWave(VOID)
{
	UINT ndevs, i;
	WAVEOUTCAPS caps;
	TCHAR buffer[256];

	ndevs = waveOutGetNumDevs();
	for (i = 0; i< ndevs; i++)
	{
		waveOutGetDevCaps(i, &caps, sizeof(WAVEOUTCAPS));
		wsprintf(buffer,_T("Device %04X: \n"), i);
		OutputDebugString(buffer);
		wsprintf(buffer,_T("  Mid     : %04X\n"), caps.wMid);
		OutputDebugString(buffer);
		wsprintf(buffer,_T("  Pid     : %04X\n"), caps.wPid);
		OutputDebugString(buffer);
		wsprintf(buffer,_T("  Version : %02X\n"), caps.vDriverVersion);
		OutputDebugString(buffer);
		wsprintf(buffer,_T("  Name    : %s\n"), caps.szPname);
		OutputDebugString(buffer);
		wsprintf(buffer,_T("  Formats : %08lX :"), caps.dwFormats);
		OutputDebugString(buffer);
		if (caps.dwFormats & WAVE_FORMAT_1M08)
		{
			wsprintf(buffer,_T("%s "), "1M08");
			OutputDebugString(buffer);
		}
		if (caps.dwFormats & WAVE_FORMAT_1S08)
		{
			wsprintf(buffer,_T("%s "), "1S08");
			OutputDebugString(buffer);
		}
		if (caps.dwFormats & WAVE_FORMAT_1M16)
		{
			wsprintf(buffer,_T("%s "), "1M16");
			OutputDebugString(buffer);
		}
		if (caps.dwFormats & WAVE_FORMAT_1S16)
		{
			wsprintf(buffer,_T("%s "), "1S16");
			OutputDebugString(buffer);
		}
		if (caps.dwFormats & WAVE_FORMAT_2M08)
		{
			wsprintf(buffer,_T("%s "), "2M08");
			OutputDebugString(buffer);
		}
		if (caps.dwFormats & WAVE_FORMAT_2S08)
		{
			wsprintf(buffer,_T("%s "), "2S08");
			OutputDebugString(buffer);
		}
		if (caps.dwFormats & WAVE_FORMAT_2M16)
		{
			wsprintf(buffer,_T("%s "), "2M16");
			OutputDebugString(buffer);
		}
		if (caps.dwFormats & WAVE_FORMAT_2S16)
		{
			wsprintf(buffer,_T("%s "), "2S16");
			OutputDebugString(buffer);
		}
		if (caps.dwFormats & WAVE_FORMAT_4M08)
		{
			wsprintf(buffer,_T("%s "), "4M08");
			OutputDebugString(buffer);
		}
		if (caps.dwFormats & WAVE_FORMAT_4S08)
		{
			wsprintf(buffer,_T("%s "), "4S08");
			OutputDebugString(buffer);
		}
		if (caps.dwFormats & WAVE_FORMAT_4M16)
		{
			wsprintf(buffer,_T("%s "), "4M16");
			OutputDebugString(buffer);
		}
		if (caps.dwFormats & WAVE_FORMAT_4S16)
		{
			wsprintf(buffer,_T("%s "), "4S16");
			OutputDebugString(buffer);
		}
		wsprintf(buffer,_T("\n"));
		OutputDebugString(buffer);
		wsprintf(buffer,_T("  Channels: %04X\n"), caps.wChannels);
		OutputDebugString(buffer);
		wsprintf(buffer,_T("  Support : %08lX :"), caps.dwSupport);
		OutputDebugString(buffer);
		if (caps.dwSupport & WAVECAPS_LRVOLUME)
		{
			wsprintf(buffer,_T("%s "), "LRVOLUME");
			OutputDebugString(buffer);
		}
		if (caps.dwSupport & WAVECAPS_PITCH)
		{
			wsprintf(buffer,_T("%s "), "PITCH");
			OutputDebugString(buffer);
		}
		if (caps.dwSupport & WAVECAPS_PLAYBACKRATE)
		{
			wsprintf(buffer,_T("%s "), "PLAYBACKRATE");
			OutputDebugString(buffer);
		}
		if (caps.dwSupport & WAVECAPS_SYNC)
		{
			wsprintf(buffer,_T("%s "), "SYNC");
			OutputDebugString(buffer);
		}
		if (caps.dwSupport & WAVECAPS_VOLUME)
		{
			wsprintf(buffer,_T("%s "), "VOLUME");
			OutputDebugString(buffer);
		}
		if (caps.dwSupport & WAVECAPS_SAMPLEACCURATE)
		{
			wsprintf(buffer,_T("%s "), "SAMPLEACCURATE");
			OutputDebugString(buffer);
		}
		wsprintf(buffer,_T("\n"));
		OutputDebugString(buffer);
	}
}
#endif

//
// Allocate a WS structure and memory for a wave of dwSamples samples (from API doc)
//
static BOOL AllocateWave(WS *wave, DWORD dwSamples)
{
    wave->hData = GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE, dwSamples ); 
    if (!wave->hData) 
        return FALSE; 

	wave->lpData = GlobalLock(wave->hData);
    if (wave->lpData == NULL) 
    { 
        GlobalFree(wave->hData); 
        return FALSE; 
    } 
    wave->hWaveHdr = GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE, (DWORD) sizeof(WAVEHDR)); 
    if (wave->hWaveHdr == NULL) 
    { 
        GlobalUnlock(wave->hData); 
        GlobalFree(wave->hData); 
        return FALSE; 
    } 
 
    wave->lpWaveHdr = (LPWAVEHDR) GlobalLock(wave->hWaveHdr); 
    if (wave->lpWaveHdr == NULL) 
    { 
        GlobalUnlock(wave->hData); 
        GlobalFree(wave->hData); 
        GlobalFree(wave->hWaveHdr); 
        return FALSE; 
    } 

    // setup block
	wave->lpWaveHdr->lpData = wave->lpData; 
    wave->lpWaveHdr->dwBufferLength = dwSamples; 
    wave->lpWaveHdr->dwFlags = 0; 
    wave->lpWaveHdr->dwLoops = 0; 
	wave->lpWaveHdr->lpNext = NULL;

	return TRUE;
}

//
// Free the structure and memory of a wave
//
static VOID FreeWave(WS *wave)
{
	if (wave->lpWaveHdr != NULL)
	{
		GlobalUnlock(wave->hWaveHdr);
		wave->lpWaveHdr = NULL;
	}
	if (wave->hWaveHdr != NULL)
	{
		GlobalFree(wave->hWaveHdr);
		wave->hWaveHdr = NULL;
	}
	if (wave->lpData != NULL)
	{
		GlobalUnlock(wave->hData);
		wave->lpData = NULL;
	}
	if (wave->hData != NULL)
	{
		GlobalFree(wave->hData);
		wave->hData = NULL;
	}
}

//################
//#
//#    Public functions
//#
//################

//
// Perform some sound initialization
//
BOOL InitWaves(VOID) 
{ 
    WAVEFORMATEX  pFormat; 

#if defined DEBUG_SND
	IdentifyWave();			// Dump info for devices
#endif

	bWaveOk = FALSE;

	SetSnRef();				// Create the reference sinus

	// Fill the format
	pFormat.wFormatTag = WAVE_FORMAT_PCM; 
	pFormat.nChannels = 1; 
	pFormat.nSamplesPerSec = SAMPLEFREQ; 
	pFormat.nAvgBytesPerSec = SAMPLEFREQ; 
	pFormat.nBlockAlign = 1; 
	pFormat.wBitsPerSample = 8; 
	pFormat.cbSize = 0;

    // open a waveform device for output without window callback. 
    if (waveOutOpen(&hWaveOut, WAVE_MAPPER, &pFormat, 
                    (DWORD) 0, 0L, CALLBACK_NULL)) 
    { 
        InfoMessage(_T("Failed to open waveform output device."));
        return FALSE; 
    } 
 
    // Allocate and lock memory for the waveform data. 
	if (! AllocateWave(&wsSound12, SAMPLES))
    { 
        InfoMessage(_T("Failed to create Sound12."));
		waveOutClose(hWaveOut);
        return FALSE; 
    } 
	if (! AllocateWave(&wsSoundxx, 10*SAMPLES))
    { 
        InfoMessage(_T("Failed to create Soundxx."));
		waveOutClose(hWaveOut);
		FreeWave(&wsSound12);
        return FALSE; 
    } 

	// Setup sound
	dwCfreq = 1200L;
	SetWaveFreq(wsSound12.lpData, dwCfreq, wsSound12.lpWaveHdr->dwBufferLength);
	SetWaveFreq(wsSoundxx.lpData, dwCfreq, wsSoundxx.lpWaveHdr->dwBufferLength);

    // After allocation, set up and prepare header.
    waveOutPrepareHeader(hWaveOut, wsSound12.lpWaveHdr, sizeof(WAVEHDR)); 
    wsSound12.lpWaveHdr->dwFlags |= WHDR_BEGINLOOP | WHDR_ENDLOOP; 
    wsSound12.lpWaveHdr->dwLoops = -1; 

    waveOutPrepareHeader(hWaveOut, wsSoundxx.lpWaveHdr, sizeof(WAVEHDR)); 
    wsSoundxx.lpWaveHdr->dwFlags |= WHDR_BEGINLOOP | WHDR_ENDLOOP; 
    wsSoundxx.lpWaveHdr->dwLoops = -1; 
 
    // Now the data block can be sent to the output device.
	bWaveOk = TRUE;

	return bWaveOk;
}

//
// Prepare to stop a sound (remember the actual sound level)
//
VOID HaltWaveInit(VOID)
{
	if (bInSound12)			// for 1.2 kHz sound just leave the loop (the sound will stop in at most .1 sec)
	{
		waveOutBreakLoop(hWaveOut);
		bInSound12 = FALSE;
	}
	else if (!bSoundStop)	// for other sound prepare a fade out
	{
		waveOutGetVolume(hWaveOut, (DWORD *) &VolumeInit);
		Volume.lw = VolumeInit.lw & 0xF000;
		waveOutSetVolume(hWaveOut, Volume.dw);
		bSoundStop = TRUE;
	}
}

//
// Really stop a sound (lower sound level and at 0 stop it)
//   called from main emulation loop
//
VOID HaltWaveDo(VOID)
{
	if (Volume.lw != 0)
	{
		Volume.lw -= 0x1000;
		Volume.hw = Volume.lw;
		waveOutSetVolume(hWaveOut, Volume.dw);
	}
	else
	{
		waveOutPause(hWaveOut);
		waveOutReset(hWaveOut);
		waveOutSetVolume(hWaveOut, VolumeInit.dw);
		bSoundStop = FALSE;
		bInSoundxx = FALSE;
		bInSound12 = FALSE;
	}
}

//
// Play the 1.2 kHz sound (if not already playing)
//
VOID PlayWave12(VOID) 
{
	if (!bInSound12)
	{
		waveOutWrite(hWaveOut, wsSound12.lpWaveHdr, sizeof(WAVEHDR)); 
		bInSound12 = TRUE;
	}
} 

//
// Play a sound at dwFreq Hz (if not already playing)
// create the samples if neccessary
//
VOID PlayWavexx(DWORD dwFreq) 
{
	if (dwFreq != dwCfreq && !bInSoundxx)
	{
		dwCfreq = dwFreq;
		SetWaveFreq(wsSoundxx.lpData, dwCfreq, wsSoundxx.lpWaveHdr->dwBufferLength);
	}
	if (!bInSoundxx)
	{
		waveOutWrite(hWaveOut, wsSoundxx.lpWaveHdr, sizeof(WAVEHDR)); 
		bInSoundxx = TRUE;
	}
} 

//
// Stop sound and deallocate all
//
BOOL CloseWaves(VOID)
{
	if (bWaveOk)
	{
		waveOutBreakLoop(hWaveOut);
		waveOutReset(hWaveOut);
		waveOutUnprepareHeader(hWaveOut, wsSound12.lpWaveHdr, sizeof(WAVEHDR)); 
		waveOutUnprepareHeader(hWaveOut, wsSoundxx.lpWaveHdr, sizeof(WAVEHDR)); 
		waveOutClose(hWaveOut);
		FreeWave(&wsSound12);
		FreeWave(&wsSoundxx);
		bWaveOk = FALSE;
	}
	return bWaveOk;
}
